"""Azure Blob Storage implementation."""

from pathlib import Path
from typing import Generator, Optional, Union

from genai_bench.auth.storage_auth_provider import StorageAuthProvider
from genai_bench.logging import init_logger
from genai_bench.storage.base import BaseStorage

logger = init_logger(__name__)


class AzureBlobStorage(BaseStorage):
    """Azure Blob Storage implementation."""

    def __init__(self, auth: StorageAuthProvider, **kwargs):
        """Initialize Azure Blob Storage.

        Args:
            auth: Storage authentication provider
            **kwargs: Additional configuration
        """
        if auth.get_storage_type() != "azure":
            raise ValueError("Auth provider must be for Azure")

        self.auth = auth
        self.config = auth.get_client_config()

        # Lazy import to avoid dependency if not using Azure
        try:
            from azure.core.exceptions import ResourceNotFoundError
            from azure.storage.blob import BlobServiceClient
        except ImportError as e:
            raise ImportError(
                "azure-storage-blob is required for Azure Blob Storage. "
                "Install it with: pip install azure-storage-blob"
            ) from e

        self.BlobServiceClient = BlobServiceClient
        self.ResourceNotFoundError = ResourceNotFoundError

        # Create blob service client
        self.client = self._create_client()

    def _create_client(self):
        """Create Azure Blob service client with appropriate credentials."""
        config = self.config

        if "connection_string" in config:
            # Use connection string
            return self.BlobServiceClient.from_connection_string(
                config["connection_string"]
            )
        elif "account_name" in config:
            # Build account URL
            account_url = f"https://{config['account_name']}.blob.core.windows.net"

            if "account_key" in config:
                # Use account key
                return self.BlobServiceClient(
                    account_url=account_url, credential=config["account_key"]
                )
            elif "sas_token" in config:
                # Use SAS token
                return self.BlobServiceClient(
                    account_url=account_url, credential=config["sas_token"]
                )
            elif config.get("use_azure_ad"):
                # Use Azure AD authentication
                try:
                    from azure.identity import (
                        ClientSecretCredential,
                        DefaultAzureCredential,
                    )
                except ImportError as e:
                    raise ImportError(
                        "azure-identity is required for Azure AD authentication. "
                        "Install it with: pip install azure-identity"
                    ) from e

                if all(
                    k in config for k in ["tenant_id", "client_id", "client_secret"]
                ):
                    # Use service principal
                    credential = ClientSecretCredential(
                        tenant_id=config["tenant_id"],
                        client_id=config["client_id"],
                        client_secret=config["client_secret"],
                    )
                else:
                    # Use default credential chain
                    credential = DefaultAzureCredential()

                return self.BlobServiceClient(
                    account_url=account_url, credential=credential
                )
            else:
                raise ValueError("No valid Azure credentials provided")
        else:
            raise ValueError("Azure account name or connection string required")

    def upload_file(
        self, local_path: Union[str, Path], remote_path: str, bucket: str, **kwargs
    ) -> None:
        """Upload a file to Azure Blob Storage.

        Args:
            local_path: Local file path
            remote_path: Blob name
            bucket: Container name
            **kwargs: Additional options (e.g., content_settings, metadata)
        """
        local_path = Path(local_path)
        if not local_path.exists():
            raise FileNotFoundError(f"Local file not found: {local_path}")

        try:
            # Get container client
            container_client = self.client.get_container_client(bucket)

            # Get blob client
            blob_client = container_client.get_blob_client(remote_path)

            # Upload file
            with open(local_path, "rb") as data:
                blob_client.upload_blob(
                    data,
                    overwrite=kwargs.get("overwrite", True),
                    content_settings=kwargs.get("content_settings"),
                    metadata=kwargs.get("metadata"),
                )

            logger.info(f"Uploaded {local_path} to azure://{bucket}/{remote_path}")

        except Exception as e:
            logger.error(f"Failed to upload {local_path}: {e}")
            raise

    def upload_folder(
        self, local_folder: Union[str, Path], bucket: str, prefix: str = "", **kwargs
    ) -> None:
        """Upload all files in a folder to Azure Blob Storage.

        Args:
            local_folder: Local folder path
            bucket: Container name
            prefix: Optional prefix for all uploaded files
            **kwargs: Additional options
        """
        local_folder = Path(local_folder)
        if not local_folder.exists() or not local_folder.is_dir():
            raise ValueError(
                f"Local folder not found or not a directory: {local_folder}"
            )

        # Upload all files in the folder
        for file_path in local_folder.rglob("*"):
            if file_path.is_file():
                # Calculate relative path
                relative_path = file_path.relative_to(local_folder)

                # Construct blob name with prefix
                if prefix:
                    blob_name = f"{prefix}/{relative_path}"
                else:
                    blob_name = str(relative_path)

                # Upload file
                self.upload_file(file_path, blob_name, bucket, **kwargs)

    def download_file(
        self, remote_path: str, local_path: Union[str, Path], bucket: str, **kwargs
    ) -> None:
        """Download a file from Azure Blob Storage.

        Args:
            remote_path: Blob name
            local_path: Local file path to save to
            bucket: Container name
            **kwargs: Additional options
        """
        local_path = Path(local_path)

        # Create parent directories if needed
        local_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            # Get container client
            container_client = self.client.get_container_client(bucket)

            # Download blob
            blob_client = container_client.get_blob_client(remote_path)

            with open(local_path, "wb") as file:
                download_stream = blob_client.download_blob()
                file.write(download_stream.readall())

            logger.info(f"Downloaded azure://{bucket}/{remote_path} to {local_path}")

        except self.ResourceNotFoundError as e:
            raise FileNotFoundError(f"Blob not found: {bucket}/{remote_path}") from e
        except Exception as e:
            logger.error(f"Failed to download {remote_path}: {e}")
            raise

    def list_objects(
        self, bucket: str, prefix: Optional[str] = None, **kwargs
    ) -> Generator[str, None, None]:
        """List objects in an Azure container with optional prefix.

        Args:
            bucket: Container name
            prefix: Optional prefix to filter objects
            **kwargs: Additional options

        Yields:
            Blob names
        """
        try:
            # Get container client
            container_client = self.client.get_container_client(bucket)

            # List blobs
            list_kwargs = {}
            if prefix:
                list_kwargs["name_starts_with"] = prefix

            for blob in container_client.list_blobs(**list_kwargs):
                yield blob.name

        except Exception as e:
            logger.error(f"Failed to list objects in {bucket}: {e}")
            raise

    def delete_object(self, remote_path: str, bucket: str, **kwargs) -> None:
        """Delete an object from Azure Blob Storage.

        Args:
            remote_path: Blob name
            bucket: Container name
            **kwargs: Additional options
        """
        try:
            # Get container client
            container_client = self.client.get_container_client(bucket)

            # Delete blob
            blob_client = container_client.get_blob_client(remote_path)
            blob_client.delete_blob()

            logger.info(f"Deleted azure://{bucket}/{remote_path}")

        except self.ResourceNotFoundError:
            logger.warning(f"Blob not found (already deleted?): {bucket}/{remote_path}")
        except Exception as e:
            logger.error(f"Failed to delete {remote_path}: {e}")
            raise

    def get_storage_type(self) -> str:
        """Get the storage provider type.

        Returns:
            'azure'
        """
        return "azure"



from typing import Optional

import oci.config

from genai_bench.auth.auth_provider import AuthProvider
from genai_bench.auth.oci.instance_principal import OCIInstancePrincipalAuth
from genai_bench.auth.oci.obo_token import OCIOBOTokenAuth
from genai_bench.auth.oci.session import OCISessionAuth
from genai_bench.auth.oci.user_principal import OCIUserPrincipalAuth
from genai_bench.auth.openai.auth import OpenAIAuth
from genai_bench.auth.together.auth import TogetherAuth


class AuthFactory:
    """Factory for creating authentication providers."""

    @staticmethod
    def create_openai_auth(api_key: str) -> OpenAIAuth:
        """Create OpenAI authentication provider.

        Args:
            api_key (str): OpenAI API key

        Returns:
            OpenAIAuth: OpenAI auth provider
        """
        return OpenAIAuth(api_key=api_key)

    @staticmethod
    def create_together_auth(api_key: str) -> TogetherAuth:
        """Create Together authentication provider.

        Args:
            api_key (str): Together API key

        Returns:
            TogetherAuth: OpenAI auth provider
        """
        return TogetherAuth(api_key=api_key)

    @staticmethod
    def create_oci_auth(
        auth_type: str,
        config_path: Optional[str] = oci.config.DEFAULT_LOCATION,
        profile: Optional[str] = oci.config.DEFAULT_PROFILE,
        token: Optional[str] = None,
        region: Optional[str] = None,
    ) -> AuthProvider:
        """Create OCI authentication provider.

        Args:
            auth_type (str): Type of OCI auth. One of: 'user_principal',
            'instance_principal', 'security_token', or 'instance_obo_user'

            config_path (Optional[str]): Path to OCI config file. Required for
            user_principal and session auth.

            profile (Optional[str]): Profile name in config.
            Only used with user_principal and session auth. Defaults to "DEFAULT".

            token (Optional[str]): token string. Required for instance_obo_user auth.
            region (Optional[str]): OCI region. Required for instance_obo_user auth.

        Returns:
            AuthProvider: OCI auth provider

        Raises:
            ValueError: If auth_type is invalid or required args are missing
        """
        # Validate auth_type first
        valid_types = [
            "user_principal",
            "instance_principal",
            "security_token",
            "instance_obo_user",
        ]
        if auth_type not in valid_types:
            raise ValueError(
                f"Invalid auth_type: {auth_type}. Must be one of: "
                f"{', '.join(repr(t) for t in valid_types)}"
            )

        if auth_type == "instance_principal":
            return OCIInstancePrincipalAuth(security_token=token, region=region)

        if auth_type == "instance_obo_user":
            if token is None or region is None:
                raise ValueError(
                    "token and region are required for obo_token authentication"
                )
            return OCIOBOTokenAuth(token=token, region=region)

        if auth_type == "user_principal":
            return OCIUserPrincipalAuth(config_path=config_path, profile=profile)

        if auth_type == "security_token":
            return OCISessionAuth(config_path=config_path, profile=profile)

        # This code is unreachable since we validate auth_type at the start
        raise ValueError(f"Unsupported auth_type: {auth_type}")  # pragma: no cover



from locust import task

import json
import time
from typing import Any, Optional

from oci.generative_ai_inference import GenerativeAiInferenceClient
from oci.generative_ai_inference.models import (
    ChatDetails,
    ChatResult,
    DedicatedServingMode,
    GenericChatRequest,
    OnDemandServingMode,
)

from genai_bench.auth.model_auth_provider import ModelAuthProvider
from genai_bench.logging import init_logger
from genai_bench.protocol import (
    UserChatRequest,
    UserChatResponse,
    UserRequest,
    UserResponse,
)
from genai_bench.user.base_user import BaseUser

logger = init_logger(__name__)

# OCI client timeout constants
OCI_CONNECT_TIMEOUT = 60
OCI_READ_TIMEOUT = 300


class OCIGenAIUser(BaseUser):
    """User class for OCI GenAI models API with OCI authentication."""

    BACKEND_NAME = "oci-genai"
    supported_tasks = {
        "text-to-text": "chat",
    }
    host: Optional[str] = None
    auth_provider: Optional[ModelAuthProvider] = None

    def on_start(self):
        """Initialize OCI client on start."""
        super().on_start()
        if not self.auth_provider:
            raise ValueError("Auth is required for OCIGenAIUser")

        # Get config and signer from auth provider
        config = self.auth_provider.get_config()
        signer = self.auth_provider.get_credentials()

        self.client = GenerativeAiInferenceClient(
            config=config,
            signer=signer,
            service_endpoint=self.host,
            timeout=(OCI_CONNECT_TIMEOUT, OCI_READ_TIMEOUT),
        )
        logger.debug("Generative AI Inference Client initialized.")

    def send_request(
        self,
        endpoint: str,
        payload: Any,
        parse_strategy: Any,
        num_prefill_tokens: Optional[int] = None,
    ) -> UserResponse:
        logger.debug(f"Sending request with payload: {payload}")
        try:
            start_time = time.monotonic()
            response = self.client.chat(payload)
            non_stream_post_end_time = time.monotonic()
            if response.status == 200:
                metrics_response = parse_strategy(
                    payload,
                    response,
                    start_time,
                    num_prefill_tokens,
                    non_stream_post_end_time,
                )
            else:
                logger.warning(
                    f"Received error status-code: {response.status} "
                    f"RequestId: {response.request_id} "
                    f"Response: {response.response}"
                )
                metrics_response = UserResponse(
                    status_code=response.status,
                    error_message="Request Failed",
                )
            self.collect_metrics(metrics_response, endpoint)
            return metrics_response
        except Exception as e:
            logger.warning(f"Error: {e}")
            return UserResponse(
                status_code=500,
                error_message=str(e),
                num_prefill_tokens=num_prefill_tokens or 0,
            )

    @task
    def chat(self):
        """Send a chat completion request using OCI GenAI Service format."""
        user_request = self.sample()

        if not isinstance(user_request, UserChatRequest):
            raise AttributeError(
                f"Expected UserChatRequest for OCIGenAIUser.chat, got "
                f"{type(user_request)}"
            )

        compartment_id = self.get_compartment_id(user_request)
        serving_mode = self.get_serving_mode(user_request)

        # Construct chat request for OCI GenAI Service format using GENERIC API format
        messages = self.build_messages(user_request)

        chat_request = GenericChatRequest(
            api_format="GENERIC",
            messages=messages,
            max_tokens=user_request.max_tokens,
            is_stream=True,
            stream_options={"isIncludeUsage": True},
            temperature=user_request.additional_request_params.get("temperature", 0.75),
            top_p=user_request.additional_request_params.get("top_p", 0.7),
            top_k=user_request.additional_request_params.get("top_k", 1),
            frequency_penalty=user_request.additional_request_params.get(
                "frequency_penalty", None
            ),
            presence_penalty=user_request.additional_request_params.get(
                "presence_penalty", None
            ),
            num_generations=1,
        )

        # Define payload with compartment ID and serving mode
        chat_detail = ChatDetails(
            compartment_id=compartment_id,
            serving_mode=serving_mode,
            chat_request=chat_request,
        )

        return self.send_request(
            endpoint="chat",
            payload=chat_detail,
            parse_strategy=self.parse_chat_response,
            num_prefill_tokens=user_request.num_prefill_tokens,
        )

    def build_messages(self, user_request: UserChatRequest) -> list:
        """Build messages array in OCI GenAI format."""
        messages = []

        # Add system message if provided
        system_message = user_request.additional_request_params.get("system_message")
        if system_message:
            messages.append(
                {
                    "role": "SYSTEM",
                    "content": [{"text": system_message, "type": "TEXT"}],
                }
            )

        # Add conversation history if provided
        chat_history = user_request.additional_request_params.get("chat_history", [])
        for msg in chat_history:
            # Convert from OpenAI format to OCI format if needed
            if isinstance(msg.get("content"), str):
                # Convert simple string content to OCI format
                oci_msg = {
                    "role": msg["role"].upper(),
                    "content": [{"text": msg["content"], "type": "TEXT"}],
                }
            else:
                # Assume it's already in OCI format
                oci_msg = msg
            messages.append(oci_msg)

        # Add current user message
        messages.append(
            {"role": "USER", "content": [{"text": user_request.prompt, "type": "TEXT"}]}
        )

        return messages

    def get_compartment_id(self, user_request: UserRequest):
        compartment_id = user_request.additional_request_params.get("compartmentId")
        if not compartment_id:
            raise ValueError("compartmentId missing in additional request params")
        return compartment_id

    def get_serving_mode(self, user_request: UserRequest) -> Any:
        params = user_request.additional_request_params
        model_id = user_request.model
        serving_type = params.get("servingType", "ON_DEMAND")

        if serving_type == "DEDICATED":
            endpoint_id = params.get("endpointId")
            if not endpoint_id:
                raise ValueError(
                    "endpointId must be provided for DEDICATED servingType"
                )
            logger.debug(
                f"Using DedicatedServingMode {serving_type} with "
                f"endpoint ID: {endpoint_id}"
            )
            return DedicatedServingMode(endpoint_id=endpoint_id)
        else:
            logger.debug(
                f"Using OnDemandServingMode {serving_type} with model ID: {model_id}"
            )
            return OnDemandServingMode(model_id=model_id)

    def parse_chat_response(
        self,
        request: ChatDetails,
        response: ChatResult,
        start_time: float,
        num_prefill_tokens: int,
        _: float,
    ) -> UserResponse:
        """
        Parses the streaming response from OCI GenAI Service using OCI's format.

        Args:
            request (ChatDetails): OCI GenAI Chat request.
            response (ChatResult): The streaming response from the API.
            start_time (float): Timestamp of request initiation.
            num_prefill_tokens (int): Number of tokens in the prompt.
            _ (float): Placeholder for unused variable.

        Returns:
            UserResponse: Parsed response in the UserResponse format.
        """
        generated_text = ""
        streaming_events_count = 0
        total_tokens = None
        prompt_tokens = None
        time_at_first_token: Optional[float] = None
        finish_reason = None
        previous_data = None
        usage_data = None

        # Iterate over each event in the streaming response
        for event in response.data.events():
            # Raw event data from the stream
            event_data = event.data.strip()
            # Parse the event data as JSON
            try:
                parsed_data = json.loads(event_data)

                # Check for usage information
                if "usage" in parsed_data:
                    usage_data = parsed_data["usage"]
                    logger.debug(f"Usage data received: {usage_data}")
                    break

                finish_reason = parsed_data.get("finishReason", None)
                if not finish_reason:
                    # Extract text content from OCI GenAI format
                    message = parsed_data.get("message", {})
                    content_array = message.get("content", [])
                    if content_array and len(content_array) > 0:
                        text_segment = content_array[0].get("text", "")
                        if text_segment:
                            # Capture the time at the first token
                            if not time_at_first_token:
                                time_at_first_token = time.monotonic()
                                logger.debug(
                                    f"First token received at: {time_at_first_token}"
                                )
                            generated_text += text_segment
                            streaming_events_count += 1  # each event contains one token
                            logger.debug(
                                f"Text: '{text_segment}', "
                                f"streaming events count: {streaming_events_count}"
                            )
                    # Track the previous data for debugging purposes
                    previous_data = parsed_data
                else:
                    # we have reached the end
                    logger.debug(
                        f"We have reached the end of the response "
                        f"with finish reason: {finish_reason}"
                    )
            except json.JSONDecodeError:
                logger.warning(
                    f"Error decoding JSON from event data: {event_data}, "
                    f"previous data: {previous_data}, "
                    f"finish reason: {finish_reason}"
                )
                continue

        # End timing for response
        end_time = time.monotonic()

        try:
            total_tokens = usage_data["totalTokens"]
            prompt_tokens = usage_data["promptTokens"]
            tokens_received = total_tokens - prompt_tokens

        except (KeyError, TypeError):
            # Fallback: use streaming events count if usage data is unavailable
            tokens_received = streaming_events_count

            logger.warning(
                f"🚨 Usage data unavailable or incomplete. "
                f"Falling back to streaming_events_count: {tokens_received}"
            )

            # Final fallback: estimate from word count if still no tokens
            if not tokens_received:
                tokens_received = len(generated_text.split())

        logger.debug(
            f"Generated text: {generated_text} \n"
            f"Time at first token: {time_at_first_token} \n"
            f"Finish reason: {finish_reason}\n"
            f"Streaming Events Count: {streaming_events_count}\n"
            f"Total Tokens (from usage): {total_tokens}\n"
            f"Prompt Tokens (from usage): {prompt_tokens}\n"
            f"Usage data: {usage_data}\n"
            f"tokens_received: {tokens_received}\n"
            f"Start Time: {start_time}\n"
            f"End Time: {end_time}\n"
        )

        # Ensure time_at_first_token is never None (fallback to end_time)
        if time_at_first_token is None:
            time_at_first_token = end_time
            logger.warning("time_at_first_token was None, using end_time as fallback")

        return UserChatResponse(
            status_code=200,
            generated_text=generated_text,
            tokens_received=tokens_received,
            time_at_first_token=time_at_first_token,
            num_prefill_tokens=num_prefill_tokens,
            start_time=start_time,
            end_time=end_time,
        )



from typing import Any, Dict, Optional

import oci

from genai_bench.auth.auth_provider import AuthProvider


class OCIInstancePrincipalAuth(AuthProvider):
    """OCI Authentication Provider using instance principal."""

    def __init__(
        self, security_token: Optional[str] = None, region: Optional[str] = None
    ):
        """Initialize OCI Instance Principal Auth Provider.

        Args:
            security_token (Optional[str]): Security token for authentication
            region (Optional[str]): OCI region
        """
        self._signer = None
        self.region = region

    def get_config(self) -> Dict[str, Any]:
        """Get OCI configuration.

        Returns:
            Dict[str, Any]: OCI configuration dictionary with region and tenancy

        Raises:
            Exception: If instance principal auth is not configured
        """
        signer = self.get_credentials()
        config = {"region": signer.region, "tenancy": signer.tenancy_id}
        if self.region:
            config["region"] = self.region
        return config

    def get_credentials(self) -> oci.signer.AbstractBaseSigner:
        """Get OCI instance principal signer.

        Returns:
            oci.signer.AbstractBaseSigner: OCI instance principal signer

        Raises:
            Exception: If instance principal auth is not configured
        """
        if self._signer is None:
            self._signer = oci.auth.signers.InstancePrincipalsSecurityTokenSigner()
        return self._signer



"""Customized user for OpenAI backends."""

from locust import task

import json
import random
import time
from typing import Any, Callable, Dict, Optional

import requests
from requests import Response

from genai_bench.auth.model_auth_provider import ModelAuthProvider
from genai_bench.logging import init_logger
from genai_bench.protocol import (
    UserChatRequest,
    UserChatResponse,
    UserEmbeddingRequest,
    UserImageChatRequest,
    UserResponse,
)
from genai_bench.user.base_user import BaseUser

logger = init_logger(__name__)


class OpenAIUser(BaseUser):
    BACKEND_NAME = "openai"
    supported_tasks = {
        "text-to-text": "chat",
        "image-text-to-text": "chat",
        "text-to-embeddings": "embeddings",
        # Future support can be added here
    }

    host: Optional[str] = None
    auth_provider: Optional[ModelAuthProvider] = None
    headers = None

    def on_start(self):
        if not self.host or not self.auth_provider:
            raise ValueError("API key and base must be set for OpenAIUser.")
        auth_headers = self.auth_provider.get_headers()
        self.headers = {
            **auth_headers,
            "Content-Type": "application/json",
        }
        self.api_backend = getattr(self, "api_backend", self.BACKEND_NAME)
        super().on_start()

    @task
    def chat(self):
        endpoint = "/v1/chat/completions"
        user_request = self.sample()

        if not isinstance(user_request, UserChatRequest):
            raise AttributeError(
                f"user_request should be of type "
                f"UserChatRequest for OpenAIUser.chat, got "
                f"{type(user_request)}"
            )

        if isinstance(user_request, UserImageChatRequest):
            text_content = [{"type": "text", "text": user_request.prompt}]
            image_content = [
                {
                    "type": "image_url",
                    "image_url": {"url": image},
                }
                for image in user_request.image_content
            ]
            content = text_content + image_content
        else:
            # Backward compatibility for vLLM versions prior to v0.5.1.
            # OpenAI API used a different text prompt format before
            # multi-modality model support.
            content = user_request.prompt

        payload = {
            "model": user_request.model,
            "messages": [
                {
                    "role": "user",
                    "content": content,
                }
            ],
            "max_tokens": user_request.max_tokens,
            "temperature": user_request.additional_request_params.get(
                "temperature", 0.0
            ),
            "stream": True,
            "stream_options": {
                "include_usage": True,
            },
            **user_request.additional_request_params,
        }

        # Conditionally add ignore_eos for vLLM and SGLang backends
        if self.api_backend in ["vllm", "sglang"]:
            payload.setdefault("ignore_eos", bool(user_request.max_tokens))
        else:
            # Remove ignore_eos for OpenAI backend, as it is not supported
            payload.pop("ignore_eos", None)

        self.send_request(
            True,
            endpoint,
            payload,
            self.parse_chat_response,
            user_request.num_prefill_tokens,
        )

    @task
    def embeddings(self):
        endpoint = "/v1/embeddings"

        user_request = self.sample()

        if not isinstance(user_request, UserEmbeddingRequest):
            raise AttributeError(
                f"user_request should be of type "
                f"UserEmbeddingRequest for OpenAIUser."
                f"embeddings, got {type(user_request)}"
            )

        random.shuffle(user_request.documents)
        payload = {
            "model": user_request.model,
            "input": user_request.documents,
            "encoding_format": user_request.additional_request_params.get(
                "encoding_format", "float"
            ),
            **user_request.additional_request_params,
        }
        self.send_request(False, endpoint, payload, self.parse_embedding_response)

    def send_request(
        self,
        stream: bool,
        endpoint: str,
        payload: Dict[str, Any],
        parse_strategy: Callable[..., UserResponse],
        num_prefill_tokens: Optional[int] = None,
    ) -> UserResponse:
        """
        Sends a POST request, handling both streaming and non-streaming
        responses.

        Args:
            endpoint (str): The API endpoint.
            payload (Dict[str, Any]): The JSON payload for the request.
            stream (bool): Whether to stream the response.
            parse_strategy (Callable[[Response, float], UserResponse]):
                The function to parse the response.
            num_prefill_tokens (Optional[int]): The num of tokens in the
                prefill/prompt. Only need for streaming requests.

        Returns:
            UserResponse: A response object containing status and metrics data.
        """
        response = None

        try:
            start_time = time.monotonic()
            response = requests.post(
                url=f"{self.host}{endpoint}",
                json=payload,
                stream=stream,
                headers=self.headers,
            )
            non_stream_post_end_time = time.monotonic()

            if response.status_code == 200:
                metrics_response = parse_strategy(
                    response,
                    start_time,
                    num_prefill_tokens,
                    non_stream_post_end_time,
                )
            else:
                metrics_response = UserResponse(
                    status_code=response.status_code,
                    error_message=response.text,
                )
        except requests.exceptions.ConnectionError as e:
            metrics_response = UserResponse(
                status_code=503, error_message=f"Connection error: {e}"
            )
        except requests.exceptions.Timeout as e:
            metrics_response = UserResponse(
                status_code=408, error_message=f"Request timed out: {e}"
            )
        except requests.exceptions.RequestException as e:
            metrics_response = UserResponse(
                status_code=500,  # Assign a generic 500
                error_message=str(e),
            )
        finally:
            if response is not None:
                response.close()

        self.collect_metrics(metrics_response, endpoint)
        return metrics_response

    def parse_chat_response(
        self,
        response: Response,
        start_time: float,
        num_prefill_tokens: int,
        _: float,
    ) -> UserResponse:
        """
        Parses a streaming response.

        Args:
            response (Response): The response object.
            start_time (float): The time when the request was started.
            num_prefill_tokens (int): The num of tokens in the prefill/prompt.
            _ (float): Placeholder for an unused var, to keep parse_*_response
                have the same interface.

        Returns:
            UserChatResponse: A response object with metrics and generated text.
        """
        stream_chunk_prefix = "data: "
        end_chunk = b"[DONE]"

        generated_text = ""
        tokens_received = 0
        time_at_first_token = None
        finish_reason = None
        previous_data = None
        num_prompt_tokens = None
        for chunk in response.iter_lines(chunk_size=None):
            # Caution: Adding logs here can make debug mode unusable.
            chunk = chunk.strip()

            if not chunk:
                continue

            chunk = chunk[len(stream_chunk_prefix) :]
            if chunk == end_chunk:
                break
            data = json.loads(chunk)

            # Handle streaming error response as OpenAI API server handles it
            # differently. Some might return 200 first and generate error response
            # later in the chunk
            if data.get("error") is not None:
                return UserResponse(
                    status_code=data["error"].get("code", -1),
                    error_message=data["error"].get(
                        "message", "Unknown error, please check server logs"
                    ),
                )

            # Standard OpenAI API streams include "finish_reason"
            # in the second-to-last chunk,
            # followed by "usage" in the final chunk,
            # which does not contain "finish_reason"
            if (
                not data["choices"]
                and finish_reason
                and "usage" in data
                and data["usage"]
            ):
                num_prefill_tokens, num_prompt_tokens, tokens_received = (
                    self._get_usage_info(data, num_prefill_tokens)
                )
                # Additional check for time_at_first_token when the response is
                # too short
                if not time_at_first_token:
                    tokens_received = data["usage"].get("completion_tokens", 0)
                    if tokens_received > 1:
                        logger.warning(
                            f"🚨🚨🚨 The first chunk the server returned "
                            f"has >1 tokens: {tokens_received}. It will "
                            f"affect the accuracy of time_at_first_token!"
                        )
                        time_at_first_token = time.monotonic()
                    else:
                        raise Exception("Invalid Response")
                break

            try:
                delta = data["choices"][0]["delta"]
                content = delta.get("content") or delta.get("reasoning_content")
                usage = delta.get("usage")

                if usage:
                    tokens_received = usage["completion_tokens"]
                if content:
                    if not time_at_first_token:
                        if tokens_received > 1:
                            logger.warning(
                                f"🚨🚨🚨 The first chunk the server returned "
                                f"has >1 tokens: {tokens_received}. It will "
                                f"affect the accuracy of time_at_first_token!"
                            )
                        time_at_first_token = time.monotonic()
                    generated_text += content

                finish_reason = data["choices"][0].get("finish_reason", None)

                # SGLang v0.4.3 to v0.4.7 has finish_reason and usage
                # in the last chunk
                if finish_reason and "usage" in data and data["usage"]:
                    num_prefill_tokens, num_prompt_tokens, tokens_received = (
                        self._get_usage_info(data, num_prefill_tokens)
                    )
                    break

            except (IndexError, KeyError) as e:
                logger.warning(
                    f"Error processing chunk: {e}, data: {data}, "
                    f"previous_data: {previous_data}, "
                    f"finish_reason: {finish_reason}, skipping"
                )

            previous_data = data

        end_time = time.monotonic()
        logger.debug(
            f"Generated text: {generated_text} \n"
            f"Time at first token: {time_at_first_token} \n"
            f"Finish reason: {finish_reason}\n"
            f"Prompt Tokens: {num_prompt_tokens} \n"
            f"Completion Tokens: {tokens_received}\n"
            f"Start Time: {start_time}\n"
            f"End Time: {end_time}"
        )

        if not tokens_received:
            tokens_received = self.environment.sampler.get_token_length(
                generated_text, add_special_tokens=False
            )
            logger.warning(
                "🚨🚨🚨 There is no usage info returned from the model "
                "server. Estimated tokens_received based on the model "
                "tokenizer."
            )
        return UserChatResponse(
            status_code=200,
            generated_text=generated_text,
            tokens_received=tokens_received,
            time_at_first_token=time_at_first_token,
            num_prefill_tokens=num_prefill_tokens,
            start_time=start_time,
            end_time=end_time,
        )

    @staticmethod
    def _get_usage_info(data, num_prefill_tokens):
        num_prompt_tokens = data["usage"]["prompt_tokens"]
        tokens_received = data["usage"]["completion_tokens"]
        # For vision task
        if num_prefill_tokens is None:
            # use num_prompt_tokens as prefill to cover image tokens
            num_prefill_tokens = num_prompt_tokens
        if abs(num_prompt_tokens - num_prefill_tokens) >= 50:
            logger.warning(
                f"Significant difference detected in prompt tokens: "
                f"The number of prompt tokens processed by the model "
                f"server ({num_prompt_tokens}) differs from the number "
                f"of prefill tokens returned by the sampler "
                f"({num_prefill_tokens}) by "
                f"{abs(num_prompt_tokens - num_prefill_tokens)} tokens."
            )
        return num_prefill_tokens, num_prompt_tokens, tokens_received

    @staticmethod
    def parse_embedding_response(
        response: Response, start_time: float, _: Optional[int], end_time: float
    ) -> UserResponse:
        """
        Parses a non-streaming response.

        Args:
            response (Response): The response object.
            start_time (float): The time when the request was started.
            _ (Optional[int]): Placeholder for an unused var, to keep
                parse_*_response have the same interface.
            end_time(float): The time when the request was finished.

        Returns:
            UserResponse: A response object with metrics.
        """

        data = response.json()
        num_prompt_tokens = data["usage"]["prompt_tokens"]

        return UserResponse(
            status_code=200,
            start_time=start_time,
            end_time=end_time,
            time_at_first_token=end_time,
            num_prefill_tokens=num_prompt_tokens,
        )



"""Time unit conversion utilities for genai-bench metrics."""

import re
from typing import Any, Dict, List, Optional


class TimeUnitConverter:
    """Utility class for converting time metrics between different units."""

    # Fields that contain time/latency values (in seconds internally)
    LATENCY_FIELDS = {"ttft", "tpot", "e2e_latency", "output_latency"}

    # Stats fields that contain time statistics
    STATS_KEYS = {
        "min",
        "max",
        "mean",
        "stddev",
        "sum",
        "p25",
        "p50",
        "p75",
        "p90",
        "p95",
        "p99",
    }

    @staticmethod
    def convert_value(
        value: Optional[float], from_unit: str, to_unit: str
    ) -> Optional[float]:
        """
        Convert a time value between units.

        Args:
            value: The time value to convert (can be None)
            from_unit: Source unit ('s' or 'ms')
            to_unit: Target unit ('s' or 'ms')

        Returns:
            Converted value or None if input was None
        """
        if value is None:
            return None

        if from_unit == to_unit:
            return value

        if from_unit == "s" and to_unit == "ms":
            return value * 1000
        elif from_unit == "ms" and to_unit == "s":
            return value / 1000
        else:
            raise ValueError(f"Unsupported unit conversion: {from_unit} -> {to_unit}")

    @classmethod
    def convert_metrics_dict(
        cls, metrics_dict: Dict[str, Any], to_unit: str, from_unit: str = "s"
    ) -> Dict[str, Any]:
        """
        Convert all time values in a metrics dictionary.

        Args:
            metrics_dict: Dictionary containing metrics data
            to_unit: Target unit ('s' or 'ms')

        Returns:
            New dictionary with converted time values
        """
        if to_unit == from_unit:
            return metrics_dict  # No conversion needed

        converted = metrics_dict.copy()

        # Convert direct latency fields
        for field in cls.LATENCY_FIELDS:
            if field in converted and converted[field] is not None:
                converted[field] = cls.convert_value(
                    converted[field], from_unit, to_unit
                )

        # Convert stats nested fields
        if "stats" in converted and isinstance(converted["stats"], dict):
            converted["stats"] = converted["stats"].copy()
            for field in cls.LATENCY_FIELDS:
                if field in converted["stats"] and isinstance(
                    converted["stats"][field], dict
                ):
                    stats_obj = converted["stats"][field].copy()
                    for stat_key in cls.STATS_KEYS:
                        if stat_key in stats_obj:
                            stats_obj[stat_key] = cls.convert_value(
                                stats_obj[stat_key], from_unit, to_unit
                            )
                    converted["stats"][field] = stats_obj

        return converted

    @classmethod
    def convert_metrics_list(
        cls, metrics_list: List[Dict[str, Any]], to_unit: str, from_unit: str = "s"
    ) -> List[Dict[str, Any]]:
        """
        Convert time values in a list of metrics dictionaries.

        Args:
            metrics_list: List of dictionaries containing metrics data
            to_unit: Target unit ('s' or 'ms')

        Returns:
            New list with converted time values
        """
        return [
            cls.convert_metrics_dict(metrics_dict, to_unit, from_unit)
            for metrics_dict in metrics_list
        ]

    @staticmethod
    def get_unit_label(base_label: str, unit: str) -> str:
        """
        Update a label with the appropriate time unit.

        Args:
            base_label: Original label (may contain '(s)' or '(ms)' or similar)
            unit: Target unit ('s' or 'ms')

        Returns:
            Updated label with correct unit notation
        """
        if unit == "ms":
            # Replace (s) or (seconds) with (ms)
            return re.sub(r"\((s|seconds)\)", "(ms)", base_label)
        elif unit == "s":
            # Replace (ms) or (milliseconds) with (s)
            return re.sub(r"\((ms|milliseconds)\)", "(s)", base_label)
        return base_label

    @staticmethod
    def validate_unit(unit: str) -> str:
        """
        Validate and normalize a time unit string.

        Args:
            unit: Time unit string

        Returns:
            Normalized unit string

        Raises:
            ValueError: If unit is not supported
        """
        unit = unit.lower().strip()
        if unit in ["s", "sec", "second", "seconds"]:
            return "s"
        elif unit in ["ms", "millisecond", "milliseconds"]:
            return "ms"
        else:
            raise ValueError(f"Unsupported time unit: {unit}. Supported units: s, ms")

    @classmethod
    def is_latency_field(cls, field_name: str) -> bool:
        """Check if a field name represents a latency metric."""
        return any(latency_field in field_name for latency_field in cls.LATENCY_FIELDS)



"""AWS S3 storage implementation."""

from pathlib import Path
from typing import Generator, Optional, Union

from genai_bench.auth.storage_auth_provider import StorageAuthProvider
from genai_bench.logging import init_logger
from genai_bench.storage.base import BaseStorage

logger = init_logger(__name__)


class AWSS3Storage(BaseStorage):
    """AWS S3 storage implementation."""

    def __init__(self, auth: StorageAuthProvider, **kwargs):
        """Initialize AWS S3 storage.

        Args:
            auth: Storage authentication provider
            **kwargs: Additional configuration
        """
        if auth.get_storage_type() != "aws":
            raise ValueError("Auth provider must be for AWS")

        self.auth = auth
        self.config = auth.get_client_config()

        # Lazy import to avoid dependency if not using AWS
        try:
            import boto3
            from botocore.exceptions import NoCredentialsError
        except ImportError as e:
            raise ImportError(
                "boto3 is required for AWS S3 storage. "
                "Install it with: pip install boto3"
            ) from e

        self.boto3 = boto3
        self.NoCredentialsError = NoCredentialsError

        # Create S3 client
        self.client = self._create_client()

    def _create_client(self):
        """Create S3 client with appropriate credentials."""
        client_kwargs = {}

        # Add region if provided
        if "region_name" in self.config:
            client_kwargs["region_name"] = self.config["region_name"]

        # Add profile if provided
        if "profile_name" in self.config:
            # Use session with profile
            session = self.boto3.Session(profile_name=self.config["profile_name"])
            return session.client("s3", **client_kwargs)
        else:
            # Add explicit credentials if provided
            if "aws_access_key_id" in self.config:
                client_kwargs["aws_access_key_id"] = self.config["aws_access_key_id"]
            if "aws_secret_access_key" in self.config:
                client_kwargs["aws_secret_access_key"] = self.config[
                    "aws_secret_access_key"
                ]
            if "aws_session_token" in self.config:
                client_kwargs["aws_session_token"] = self.config["aws_session_token"]

            return self.boto3.client("s3", **client_kwargs)

    def upload_file(
        self, local_path: Union[str, Path], remote_path: str, bucket: str, **kwargs
    ) -> None:
        """Upload a file to S3.

        Args:
            local_path: Local file path
            remote_path: S3 key
            bucket: S3 bucket name
            **kwargs: Additional options (e.g., ExtraArgs)
        """
        local_path = Path(local_path)
        if not local_path.exists():
            raise FileNotFoundError(f"Local file not found: {local_path}")

        try:
            # Check if multipart upload is needed (files > 100MB)
            file_size = local_path.stat().st_size
            if file_size > 100 * 1024 * 1024:  # 100MB
                # Use multipart upload
                logger.info(
                    f"Using multipart upload for large file "
                    f"({file_size / 1024 / 1024:.2f} MB)"
                )
                self._multipart_upload(local_path, remote_path, bucket, **kwargs)
            else:
                # Regular upload
                extra_args = kwargs.get("ExtraArgs", {})
                self.client.upload_file(
                    str(local_path), bucket, remote_path, ExtraArgs=extra_args
                )

            logger.info(f"Uploaded {local_path} to s3://{bucket}/{remote_path}")

        except self.NoCredentialsError as e:
            raise ValueError("AWS credentials not found") from e
        except Exception as e:
            logger.error(f"Failed to upload {local_path}: {e}")
            raise

    def _multipart_upload(
        self, local_path: Path, remote_path: str, bucket: str, **kwargs
    ) -> None:
        """Perform multipart upload for large files."""
        from boto3.s3.transfer import TransferConfig

        # Configure multipart upload
        config = TransferConfig(
            multipart_threshold=1024 * 25,  # 25MB
            max_concurrency=10,
            multipart_chunksize=1024 * 25,
            use_threads=True,
        )

        extra_args = kwargs.get("ExtraArgs", {})
        self.client.upload_file(
            str(local_path), bucket, remote_path, ExtraArgs=extra_args, Config=config
        )

    def upload_folder(
        self, local_folder: Union[str, Path], bucket: str, prefix: str = "", **kwargs
    ) -> None:
        """Upload all files in a folder to S3.

        Args:
            local_folder: Local folder path
            bucket: S3 bucket name
            prefix: Optional prefix for all uploaded files
            **kwargs: Additional options
        """
        local_folder = Path(local_folder)
        if not local_folder.exists() or not local_folder.is_dir():
            raise ValueError(
                f"Local folder not found or not a directory: {local_folder}"
            )

        # Upload all files in the folder
        for file_path in local_folder.rglob("*"):
            if file_path.is_file():
                # Calculate relative path
                relative_path = file_path.relative_to(local_folder)

                # Construct S3 key with prefix
                s3_key = f"{prefix}/{relative_path}" if prefix else str(relative_path)

                # Upload file
                self.upload_file(file_path, s3_key, bucket, **kwargs)

    def download_file(
        self, remote_path: str, local_path: Union[str, Path], bucket: str, **kwargs
    ) -> None:
        """Download a file from S3.

        Args:
            remote_path: S3 key
            local_path: Local file path to save to
            bucket: S3 bucket name
            **kwargs: Additional options
        """
        local_path = Path(local_path)

        # Create parent directories if needed
        local_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            self.client.download_file(bucket, remote_path, str(local_path))
            logger.info(f"Downloaded s3://{bucket}/{remote_path} to {local_path}")
        except self.NoCredentialsError as e:
            raise ValueError("AWS credentials not found") from e
        except Exception as e:
            logger.error(f"Failed to download {remote_path}: {e}")
            raise

    def list_objects(
        self, bucket: str, prefix: Optional[str] = None, **kwargs
    ) -> Generator[str, None, None]:
        """List objects in an S3 bucket with optional prefix.

        Args:
            bucket: S3 bucket name
            prefix: Optional prefix to filter objects
            **kwargs: Additional options

        Yields:
            Object keys
        """
        try:
            paginator = self.client.get_paginator("list_objects_v2")

            page_kwargs = {"Bucket": bucket}
            if prefix:
                page_kwargs["Prefix"] = prefix

            for page in paginator.paginate(**page_kwargs):
                if "Contents" in page:
                    for obj in page["Contents"]:
                        yield obj["Key"]

        except self.NoCredentialsError as e:
            raise ValueError("AWS credentials not found") from e
        except Exception as e:
            logger.error(f"Failed to list objects in {bucket}: {e}")
            raise

    def delete_object(self, remote_path: str, bucket: str, **kwargs) -> None:
        """Delete an object from S3.

        Args:
            remote_path: S3 key
            bucket: S3 bucket name
            **kwargs: Additional options
        """
        try:
            self.client.delete_object(Bucket=bucket, Key=remote_path)
            logger.info(f"Deleted s3://{bucket}/{remote_path}")
        except self.NoCredentialsError as e:
            raise ValueError("AWS credentials not found") from e
        except Exception as e:
            logger.error(f"Failed to delete {remote_path}: {e}")
            raise

    def get_storage_type(self) -> str:
        """Get the storage provider type.

        Returns:
            'aws'
        """
        return "aws"



"""Azure Blob Storage authentication provider."""

import os
from typing import Any, Dict, Optional

from genai_bench.auth.storage_auth_provider import StorageAuthProvider


class AzureBlobAuth(StorageAuthProvider):
    """Azure Blob Storage authentication provider."""

    def __init__(
        self,
        account_name: Optional[str] = None,
        account_key: Optional[str] = None,
        connection_string: Optional[str] = None,
        sas_token: Optional[str] = None,
        use_azure_ad: bool = False,
        tenant_id: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        """Initialize Azure Blob Storage authentication.

        Args:
            account_name: Storage account name
            account_key: Storage account key
            connection_string: Full connection string (overrides other params)
            sas_token: Shared Access Signature token
            use_azure_ad: Whether to use Azure AD authentication
            tenant_id: Azure AD tenant ID
            client_id: Azure AD client/application ID
            client_secret: Azure AD client secret
        """
        self.account_name = account_name or os.getenv("AZURE_STORAGE_ACCOUNT_NAME")
        self.account_key = account_key or os.getenv("AZURE_STORAGE_ACCOUNT_KEY")
        self.connection_string = connection_string or os.getenv(
            "AZURE_STORAGE_CONNECTION_STRING"
        )
        self.sas_token = sas_token or os.getenv("AZURE_STORAGE_SAS_TOKEN")
        self.use_azure_ad = use_azure_ad
        self.tenant_id = tenant_id or os.getenv("AZURE_TENANT_ID")
        self.client_id = client_id or os.getenv("AZURE_CLIENT_ID")
        self.client_secret = client_secret or os.getenv("AZURE_CLIENT_SECRET")

    def get_client_config(self) -> Dict[str, Any]:
        """Get Azure Blob Storage client configuration.

        Returns:
            Dict[str, Any]: Configuration for storage client
        """
        config: Dict[str, Any] = {}

        if self.connection_string:
            config["connection_string"] = self.connection_string
        else:
            if self.account_name:
                config["account_name"] = self.account_name

            if self.account_key:
                config["account_key"] = self.account_key
            elif self.sas_token:
                config["sas_token"] = self.sas_token
            elif self.use_azure_ad:
                config["use_azure_ad"] = self.use_azure_ad
                if self.tenant_id:
                    config["tenant_id"] = self.tenant_id
                if self.client_id:
                    config["client_id"] = self.client_id
                if self.client_secret:
                    config["client_secret"] = self.client_secret

        return config

    def get_credentials(self) -> Dict[str, Any]:
        """Get Azure credentials for storage operations.

        Returns:
            Dict with Azure credentials
        """
        creds: Dict[str, Any] = {}

        if self.connection_string:
            creds["connection_string"] = self.connection_string
        else:
            if self.account_name:
                creds["account_name"] = self.account_name

            if self.account_key:
                creds["account_key"] = self.account_key
            elif self.sas_token:
                creds["sas_token"] = self.sas_token
            elif self.use_azure_ad:
                creds["use_azure_ad"] = self.use_azure_ad
                if self.tenant_id:
                    creds["tenant_id"] = self.tenant_id
                if self.client_id:
                    creds["client_id"] = self.client_id
                if self.client_secret:
                    creds["client_secret"] = self.client_secret

        return creds

    def get_storage_type(self) -> str:
        """Get the storage provider type.

        Returns:
            'azure'
        """
        return "azure"



from typing import Any, Set

from datasets import Dataset as HFDataset
from datasets import DatasetDict as HFDatasetDict
from datasets import IterableDataset as HFIterableDataset
from datasets import IterableDatasetDict as HFIterableDatasetDict

from genai_bench.data.loaders.base import DatasetFormat, DatasetLoader
from genai_bench.logging import init_logger

logger = init_logger(__name__)


class ImageDatasetLoader(DatasetLoader):
    """
    This datasetLoader is responsible for loading images and prompts from a data source.
    """

    supported_formats: Set[DatasetFormat] = {
        DatasetFormat.HUGGINGFACE_HUB,
    }
    media_type = "Image"

    def _process_loaded_data(self, data: Any) -> Any:
        """Normalize HF datasets and avoid eager iteration.

        - If it's a `datasets.Dataset`, return it (supports len and __getitem__).
        - If it's a `datasets.DatasetDict`, select a split (prefer 'train' or the
          first available) and return that `Dataset`.
        - If it's a streaming dataset (`IterableDataset` or `IterableDatasetDict`),
          raise an error instructing to disable streaming.
        - Otherwise, pass-through.
        """
        if isinstance(data, HFDataset):
            return data
        if isinstance(data, HFDatasetDict):
            available_splits = list(data.keys())
            if not available_splits:
                raise ValueError(
                    "HuggingFace DatasetDict has no splits to select from."
                )
            chosen_split = "train" if "train" in data else available_splits[0]
            if len(available_splits) > 1 and chosen_split != "train":
                logger.warning(
                    "Multiple splits found %s; defaulting to '%s'. "
                    "Set config.huggingface_kwargs.split to control this.",
                    available_splits,
                    chosen_split,
                )
            return data[chosen_split]
        if isinstance(data, (HFIterableDataset, HFIterableDatasetDict)):
            raise ValueError(
                "Streaming datasets are not supported for image sampling. "
                "Load without streaming (streaming=False) and provide a concrete split."
            )
        return data



"""OCI storage authentication adapter for backward compatibility."""

from typing import Any, Dict, Optional

from genai_bench.auth.auth_provider import AuthProvider
from genai_bench.auth.storage_auth_provider import StorageAuthProvider


class OCIStorageAuthAdapter(StorageAuthProvider):
    """Adapter to use existing OCI auth providers as storage auth providers."""

    def __init__(self, oci_auth: AuthProvider):
        """Initialize OCI storage auth adapter.

        Args:
            oci_auth: Existing OCI auth provider instance
        """
        self.oci_auth = oci_auth

    def get_client_config(self) -> Dict[str, Any]:
        """Get OCI storage client configuration.

        Returns:
            Dict[str, Any]: Configuration for storage client
        """
        return {
            "auth_provider": self.oci_auth,
        }

    def get_credentials(self) -> Any:
        """Get OCI authentication credentials.

        Returns:
            The OCI auth provider instance
        """
        return self.oci_auth

    def get_storage_type(self) -> str:
        """Get the storage provider type.

        Returns:
            'oci'
        """
        return "oci"

    def get_region(self) -> Optional[str]:
        """Get the OCI region.

        Returns:
            OCI region from the auth provider
        """
        # Try to get region from the auth provider if available
        if self.oci_auth and hasattr(self.oci_auth, "region"):
            return self.oci_auth.region
        return None



import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Tuple, Type


class TextDistribution(Enum):
    """
    Distribution type for text input output sampling
    """

    NORMAL = "N"
    DETERMINISTIC = "D"
    UNIFORM = "U"


class EmbeddingDistribution(Enum):
    """
    Distribution type for embedding input sampling
    """

    EMBEDDING = "E"


class ReRankDistribution(Enum):
    """
    Distribution type for re-ranking documents as per query.
    """

    RE_RANK = "R"


class MultiModality(Enum):
    """
    Enumeration for multi modality scenario sampling
    """

    IMAGE = "I"  # I(width, height)
    VIDEO = "V"
    AUDIO = "A"


class SpecialScenario(Enum):
    """Special, non-parametric scenario types."""

    DATASET = "dataset"


class Scenario(ABC):
    """
    An abstract base class for different scenarios based on specified
    distributions or modalities.
    """

    _registry: Dict[str, Type["Scenario"]] = {}
    scenario_type: (
        TextDistribution
        | MultiModality
        | EmbeddingDistribution
        | ReRankDistribution
        | SpecialScenario
    )
    validation_pattern: str

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._registry[cls.scenario_type.value] = cls

    @abstractmethod
    def sample(self) -> Any:
        """
        Samples the number of input and output tokens based on the distribution
        and parameters.
        Or sampling different modalities specs such as
        image width and height
        """
        pass

    @abstractmethod
    def to_string(self) -> str:
        """Converts the Scenario object back into its string representation."""
        pass

    @classmethod
    @abstractmethod
    def parse(cls, params_str: str) -> "Scenario":
        """Parses scenario str to a Scenario object."""
        pass

    @classmethod
    def from_string(cls, scenario_str: str) -> "Scenario":
        """
        Factory method that creates a Scenario object from its string
        representation.
        Determines the correct subclass to instantiate using the registered map.
        """
        # Extract leading type token (supports multi-char tokens like "dataset")
        match = re.match(r"^([A-Za-z]+)", scenario_str)
        type_token = match.group(1) if match else scenario_str[0]
        cls.validate(scenario_str)
        scenario_class = cls._registry.get(type_token)
        assert (
            scenario_class is not None
        ), "scenario_class should not be None at this step"
        # Pass the parameter substring (if any) to parser
        params_str = scenario_str[len(type_token) :]
        return scenario_class.parse(params_str)

    @classmethod
    def validate(cls, scenario_str: str) -> bool:
        """
        Scenario string validation method
        Subclass requires validation_pattern to be defined
        """
        match = re.match(r"^([A-Za-z]+)", scenario_str)
        type_token = match.group(1) if match else scenario_str[0]
        scenario_class = cls._registry.get(type_token)
        if not scenario_class:
            supported = ", ".join(sorted(cls._registry.keys()))
            raise ValueError(
                f"Invalid scenario string '{scenario_str}'. Unknown type "
                f"'{type_token}'. Supported types: {supported}"
            )
        if not re.match(scenario_class.validation_pattern, scenario_str):
            raise ValueError(
                f"Invalid scenario string '{scenario_str}' for type '{type_token}'. "
                f"Expected to match pattern: {scenario_class.validation_pattern}"
            )
        return True


class DatasetScenario(Scenario):
    """
    A generic no-op scenario used to indicate dataset/direct sampling mode.
    It is registered in the scenario registry and created via its human-friendly
    alias in Scenario.from_string (e.g., "dataset").
    """

    scenario_type = SpecialScenario.DATASET
    validation_pattern = r"^dataset$"

    def sample(self) -> Any:  # pragma: no cover
        raise NotImplementedError(
            "DatasetScenario has no sampling parameters; samplers should bypass token "
            "shaping in dataset mode."
        )

    def to_string(self) -> str:
        return "dataset"

    @classmethod
    def parse(cls, params_str: str) -> "Scenario":
        return cls()


def parse_params_str(
    params_str: str,
) -> List[Tuple[int, ...]]:
    """
    Parses the parameter string into tuples of integers.

    Args:
        params_str: The parameter string to parse.

    Returns:
        A list of tuples containing integers parsed from the string.
    """
    if "/" in params_str:
        parts = params_str.split("/")
        return [tuple(map(int, part[1:-1].split(","))) for part in parts]
    else:
        return [tuple(map(int, params_str[1:-1].split(",")))]



import os
import time
from typing import Dict, List, Optional, Union

from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.text import Text

from genai_bench.logging import init_logger
from genai_bench.protocol import LiveMetricsData
from genai_bench.ui.layout import (
    create_layout,
    create_metric_panel,
    create_progress_bars,
    update_progress,
)
from genai_bench.ui.plots import (
    create_horizontal_colored_bar_chart,
    create_scatter_plot,
)

logger = init_logger(__name__)


class MinimalDashboard:
    """A minimal implementation of the Dashboard interface for scenarios without UI."""

    def __init__(self, metrics_time_unit: str = "s"):
        self.console = None
        self.layout = None
        self.metrics_time_unit = metrics_time_unit
        self._live = type(
            "MinimalDashboardLive",
            (),
            {
                "__enter__": lambda x: None,
                "__exit__": lambda x, *args: None,
            },
        )()  # a simple no-op context manager to work with dashboard.live in cli.py

    @property
    def live(self):
        return self._live

    def update_metrics_panels(
        self, live_metrics: LiveMetricsData, metrics_time_unit: str = "s"
    ):
        pass

    def update_histogram_panel(
        self, live_metrics: LiveMetricsData, metrics_time_unit: str = "s"
    ):
        pass

    def update_scatter_plot_panel(
        self, _ui_scatter_plot_metrics: Optional[List[float]], _time_unit: str = "s"
    ):
        pass

    def update_benchmark_progress_bars(self, progress_increment: float):
        pass

    def create_benchmark_progress_task(self, run_name: str):
        pass

    def update_total_progress_bars(self, total_runs: int):
        pass

    def start_run(self, run_time: int, start_time: float, max_requests_per_run: int):
        pass

    def calculate_time_based_progress(self) -> float:
        return 0.0

    def handle_single_request(
        self, live_metrics: LiveMetricsData, total_requests: int, error_code: int | None
    ):
        pass

    def reset_plot_metrics(self):
        pass

    def reset_panels(self):
        pass


class RichLiveDashboard:
    """
    A dashboard implementation leveraging the `rich` library for live updates and
    interactive UI.

    The RichLiveDashboard provides real-time metrics visualization and progress tracking
    for benchmarking scenarios. It dynamically updates panels, histograms, and scatter
    plots to display key performance data, offering an intuitive and visually appealing
    experience.

    Features:
    - **Rich UI Components**: Uses the `rich` library to create visually rich layouts
      with panels, progress bars, and charts.
    - **Live Updates**: Dynamically updates metrics, histograms, and scatter plots in
      real time.
    - **Interactive Visualization**: Supports the visualization of performance metrics
      like latency, throughput, and task progress.
    - **Custom Layouts**: Provides a structured layout to organize data effectively for
      users.

    Typical Use Case:
    - Designed for benchmarking scenarios where real-time feedback on metrics is
      essential.
    - Enables monitoring of both system performance and task progress.

    Key Panels and Visuals:
    - **Metrics Panels**: Displays input and output latency/throughput metrics.
    - **Histograms**: Visualizes latency distributions with horizontal bar charts.
    - **Scatter Plots**: Correlates throughput and latency for input and output
      phases.
    - **Progress Bars**: Tracks overall benchmarking progress and individual run
      progress.

    Initialization Parameters:
    - No parameters are required. The dashboard initializes with default layouts
      and refresh rates.

    Example:
        dashboard = RichLiveDashboard()
        with dashboard.live:
            dashboard.update_metrics_panels(live_metrics)
            dashboard.update_benchmark_progress_bars(progress_increment)
            ...

    Dependencies:
    - `rich` library for UI and visualization components.
    - `genai_bench` modules for metrics and logging.

    Notes:
    - This class assumes the availability of metrics in the form of `LiveMetricsData`.
    - If `ENABLE_UI` is disabled, consider using `MinimalDashboard` instead.
    """

    def __init__(self, metrics_time_unit: str = "s"):
        self.console: Console = Console()
        self.layout = create_layout()
        (
            self.total_progress,
            self.benchmark_progress,
            self.total_progress_task_id,
        ) = create_progress_bars()
        self.benchmark_progress_task_id: Optional[int] = None
        self.start_time: Optional[float] = None
        self.run_time: Optional[int] = None
        self.max_requests_per_run: Optional[int] = None
        self.metrics_time_unit: str = metrics_time_unit
        self.plot_metrics: Dict[str, List[float]] = {
            "ttft": [],
            "input_throughput": [],
            "output_throughput": [],
            "output_latency": [],
        }
        self.live: Live = Live(
            self.layout,
            console=self.console,
            refresh_per_second=2,
            screen=True,
        )

    def update_metrics_panels(
        self, live_metrics: LiveMetricsData, metrics_time_unit: str = "s"
    ):
        if "stats" not in live_metrics or not live_metrics["stats"]:
            return
        stats = live_metrics["stats"]
        # Handle both dict and list formats for stats
        if isinstance(stats, dict):
            input_latency_panel, input_throughput_panel = create_metric_panel(
                "Input",
                stats.get("ttft", []),
                stats.get("input_throughput", []),
                metrics_time_unit,
            )
            output_latency_panel, output_throughput_panel = create_metric_panel(
                "Output",
                stats.get("output_latency", []),
                stats.get("output_throughput", []),
                metrics_time_unit,
            )
        else:
            # If stats is a list or other format, use empty lists as fallback
            input_latency_panel, input_throughput_panel = create_metric_panel(
                "Input", [], [], metrics_time_unit
            )
            output_latency_panel, output_throughput_panel = create_metric_panel(
                "Output", [], [], metrics_time_unit
            )

        self.layout["input_throughput"].update(input_throughput_panel)
        self.layout["input_latency"].update(input_latency_panel)
        self.layout["output_throughput"].update(output_throughput_panel)
        self.layout["output_latency"].update(output_latency_panel)

    def update_histogram_panel(
        self, live_metrics: LiveMetricsData, metrics_time_unit: str = "s"
    ):
        input_latency_hist_chart = create_horizontal_colored_bar_chart(
            live_metrics["ttft"], bin_width=0.01, metrics_time_unit=metrics_time_unit
        )
        output_latency_hist_chart = create_horizontal_colored_bar_chart(
            live_metrics["output_latency"],
            bin_width=0.01,
            metrics_time_unit=metrics_time_unit,
        )

        self.layout["input_histogram"].update(
            Panel(
                input_latency_hist_chart,
                title="Input Latency Histogram",
                border_style="bold green",
                expand=True,
            )
        )
        self.layout["output_histogram"].update(
            Panel(
                output_latency_hist_chart,
                title="Output Latency Histogram",
                border_style="bold blue",
                expand=True,
            )
        )

    def update_scatter_plot_panel(
        self,
        ui_scatter_plot_metrics: Optional[List[float]],
        metrics_time_unit: str = "s",
    ):
        if not ui_scatter_plot_metrics:
            logger.info("No ui_scatter_plot_metrics collected for this run.")
            return

        ttft, output_latency, input_throughput, output_throughput = (
            ui_scatter_plot_metrics
        )
        self.plot_metrics["ttft"].append(ttft)
        self.plot_metrics["input_throughput"].append(input_throughput)
        self.plot_metrics["output_throughput"].append(output_throughput)
        self.plot_metrics["output_latency"].append(output_latency)

        input_throughput_scatter_plot = create_scatter_plot(
            self.plot_metrics["input_throughput"],
            self.plot_metrics["ttft"],
            y_unit=metrics_time_unit,
            x_unit="tokens/sec",
        )
        output_latency_scatter_plot = create_scatter_plot(
            self.plot_metrics["output_throughput"],
            self.plot_metrics["output_latency"],
            y_unit=metrics_time_unit,
            x_unit="tokens/sec",
        )

        self.layout["ttft_vs_input_throughput"].update(
            Panel(
                input_throughput_scatter_plot,
                title="Input Latency vs Input Throughput of Server",
                border_style="bold green",
                expand=True,
            )
        )
        self.layout["output_latency_vs_output_throughput"].update(
            Panel(
                output_latency_scatter_plot,
                title="Output Latency vs Output Throughput of Server",
                border_style="bold blue",
                expand=True,
            )
        )

    def update_benchmark_progress_bars(self, progress_increment: float):
        self.benchmark_progress.update(
            self.benchmark_progress_task_id, completed=progress_increment
        )
        update_progress(self.layout, self.total_progress, self.benchmark_progress)

    def create_benchmark_progress_task(self, run_name: str):
        self.benchmark_progress_task_id = self.benchmark_progress.add_task(
            run_name, total=100
        )
        update_progress(self.layout, self.total_progress, self.benchmark_progress)

    def update_total_progress_bars(self, total_runs: int):
        self.benchmark_progress.remove_task(self.benchmark_progress_task_id)
        self.total_progress.update(
            self.total_progress_task_id, advance=(1 / total_runs) * 100
        )
        update_progress(self.layout, self.total_progress, self.benchmark_progress)

    def start_run(self, run_time: int, start_time: float, max_requests_per_run: int):
        self.start_time = start_time
        self.run_time = run_time
        self.max_requests_per_run = max_requests_per_run

    def calculate_time_based_progress(self) -> float:
        assert self.start_time is not None and self.run_time is not None
        time_elapsed = time.monotonic() - self.start_time
        return min(time_elapsed / self.run_time, 1) * 100

    def handle_single_request(
        self,
        live_metrics: LiveMetricsData,
        total_requests: int,
        error_code: int | None,
    ):
        # Calculate time-based progress
        time_based_progress = self.calculate_time_based_progress()

        # Calculate request-based progress
        assert self.max_requests_per_run is not None
        request_based_progress = (
            min(total_requests / self.max_requests_per_run, 1) * 100
        )

        # Use the larger of the two progress metrics to be more accurate
        progress_increment = max(time_based_progress, request_based_progress)

        self.update_benchmark_progress_bars(progress_increment)

        # No need to update metrics panel or histogram panel when the request
        # fails
        if error_code is not None:
            return

        self.update_metrics_panels(live_metrics, self.metrics_time_unit)
        self.update_histogram_panel(live_metrics, self.metrics_time_unit)

    def reset_plot_metrics(self):
        """Reset plot metrics for each scenario."""
        self.plot_metrics = {
            "ttft": [],
            "input_throughput": [],
            "output_throughput": [],
            "output_latency": [],
        }

        self.layout["ttft_vs_input_throughput"].update(
            Panel(
                Text(),
                title="Input Latency vs Input Throughput of Server",
                border_style="bold green",
                expand=True,
            )
        )
        self.layout["output_latency_vs_output_throughput"].update(
            Panel(
                Text(),
                title="Output Latency vs Output Throughput of Server",
                border_style="bold blue",
                expand=True,
            )
        )

    def reset_panels(self):
        """Reset the histogram and live panel."""
        self.layout["input_throughput"].update(
            self._create_empty_panel("Input Throughput", "blue")
        )
        self.layout["input_latency"].update(
            self._create_empty_panel("Input Latency", "yellow")
        )
        self.layout["output_throughput"].update(
            self._create_empty_panel("Output Throughput", "blue")
        )
        self.layout["output_latency"].update(
            self._create_empty_panel("Output Latency", "yellow")
        )
        self.layout["input_histogram"].update(
            self._create_empty_panel("Input Latency Histogram", "bold green")
        )
        self.layout["output_histogram"].update(
            self._create_empty_panel("Output Latency Histogram", "bold blue")
        )

    @staticmethod
    def _create_empty_panel(title: str, border_style: str) -> Panel:
        return Panel(
            Text(),
            title=title,
            border_style=f"{border_style}",
            expand=True,
        )


Dashboard = Union[RichLiveDashboard, MinimalDashboard]


def create_dashboard(metrics_time_unit: str = "s") -> Dashboard:
    """Factory function that returns either a NoOpDashboard or RealDashboard based
    on ENABLE_UI."""
    enable_ui_str = os.getenv("ENABLE_UI", "true").lower()
    enable_ui = enable_ui_str in ("true", "1", "yes", "on")
    return (
        RichLiveDashboard(metrics_time_unit)
        if enable_ui
        else MinimalDashboard(metrics_time_unit)
    )



"""GCP Vertex AI user implementation."""

from locust import task

import json
import time
from typing import Any, Callable, Dict, List, Optional

import requests
from requests import Response

from genai_bench.auth.auth_provider import AuthProvider
from genai_bench.logging import init_logger
from genai_bench.protocol import (
    UserChatRequest,
    UserChatResponse,
    UserEmbeddingRequest,
    UserImageChatRequest,
    UserResponse,
)
from genai_bench.user.base_user import BaseUser

logger = init_logger(__name__)


class GCPVertexUser(BaseUser):
    """GCP Vertex AI API user implementation."""

    BACKEND_NAME = "gcp-vertex"

    # Task mapping
    supported_tasks = {
        "text-to-text": "chat",
        "text-to-embeddings": "embeddings",
        "image-text-to-text": "chat",  # Same method handles both text and image
    }

    host: Optional[str] = None
    auth_provider: Optional[AuthProvider] = None
    project_id: Optional[str] = None
    location: Optional[str] = None
    headers: Dict[str, str] = {}

    def on_start(self):
        """Initialize GCP Vertex AI client using auth provider."""
        if not self.auth_provider:
            raise ValueError("Auth provider not set for GCP Vertex AI")

        # Get config
        config = self.auth_provider.get_config()
        self.project_id = config.get("project_id")
        self.location = config.get("location", "us-central1")

        if not self.project_id:
            raise ValueError("Project ID is required for Vertex AI")

        # Set up authentication
        auth_type = config.get("auth_type")

        if auth_type == "api_key":
            # API key authentication
            self.headers = self.auth_provider.get_headers()
            self.headers["Content-Type"] = "application/json"
        else:
            # Service account authentication
            try:
                import google.auth
                from google.auth.transport.requests import Request
                from google.oauth2 import service_account
            except ImportError as e:
                raise ImportError(
                    "google-auth is required for Vertex AI service account auth. "
                    "Install it with: pip install google-auth"
                ) from e

            credentials_path = config.get("credentials_path")

            if credentials_path:
                # Load service account credentials
                credentials = service_account.Credentials.from_service_account_file(
                    credentials_path,
                    scopes=["https://www.googleapis.com/auth/cloud-platform"],
                )
            else:
                # Use default credentials
                credentials, _ = google.auth.default(
                    scopes=["https://www.googleapis.com/auth/cloud-platform"]
                )

            # Get access token
            auth_req = Request()
            credentials.refresh(auth_req)

            self.headers = {
                "Authorization": f"Bearer {credentials.token}",
                "Content-Type": "application/json",
            }

        logger.info(f"Initialized GCP Vertex AI client for project {self.project_id}")
        super().on_start()

    @task
    def chat(self):
        """Perform a chat request to GCP Vertex AI."""
        # Get request using sample method
        user_request = self.sample()

        if not isinstance(user_request, UserChatRequest):
            raise AttributeError(
                "user_request should be of type UserChatRequest for "
                f"GCPVertexUser.chat, got {type(user_request)}"
            )

        # Determine model family
        model_name = user_request.model
        is_gemini = "gemini" in model_name.lower()
        use_streaming = user_request.additional_request_params.get("stream", True)

        # Build URL based on model type
        if is_gemini:
            # Gemini models use generateContent endpoint
            endpoint_method = (
                "streamGenerateContent" if use_streaming else "generateContent"
            )
            endpoint = (
                f"/v1/projects/{self.project_id}/locations/{self.location}/"
                f"publishers/google/models/{model_name}:{endpoint_method}"
            )
        else:
            # Other models (e.g., text-bison) use predict endpoint
            endpoint = (
                f"/v1/projects/{self.project_id}/locations/{self.location}/"
                f"publishers/google/models/{model_name}:predict"
            )

        # Prepare request body
        request_body = self._prepare_request_body(user_request, model_name)

        # Send request
        self.send_request(
            use_streaming and is_gemini,  # Only Gemini supports streaming
            endpoint,
            request_body,
            self.parse_chat_response if is_gemini else self.parse_palm_response,
            user_request.num_prefill_tokens,
            model_name,
        )

    @task
    def embeddings(self):
        """Perform an embeddings request to GCP Vertex AI."""
        # Get request using sample method
        user_request = self.sample()

        if not isinstance(user_request, UserEmbeddingRequest):
            raise AttributeError(
                "user_request should be of type UserEmbeddingRequest for "
                f"GCPVertexUser.embeddings, got {type(user_request)}"
            )

        # Build URL for embeddings model
        model_name = user_request.model
        endpoint = (
            f"/v1/projects/{self.project_id}/locations/{self.location}/"
            f"publishers/google/models/{model_name}:predict"
        )

        # Prepare request body
        # Vertex AI expects each document as a separate instance
        request_body = {
            "instances": [{"content": doc} for doc in user_request.documents]
        }

        # Add additional params
        if user_request.additional_request_params:
            request_body["parameters"] = user_request.additional_request_params

        # Send request
        self.send_request(
            False,
            endpoint,
            request_body,
            self.parse_embedding_response,
        )

    def send_request(
        self,
        stream: bool,
        endpoint: str,
        payload: Dict[str, Any],
        parse_strategy: Callable[..., UserResponse],
        num_prefill_tokens: Optional[int] = None,
        model_name: Optional[str] = None,
    ) -> UserResponse:
        """
        Sends a POST request, handling both streaming and non-streaming responses.

        Args:
            endpoint (str): The API endpoint.
            payload (Dict[str, Any]): The JSON payload for the request.
            stream (bool): Whether to stream the response.
            parse_strategy (Callable): The function to parse the response.
            num_prefill_tokens (Optional[int]): The num of tokens in the prefill/prompt.
            model_name (Optional[str]): The model name for response parsing.

        Returns:
            UserResponse: A response object containing status and metrics data.
        """
        response = None
        base_url = f"https://{self.location}-aiplatform.googleapis.com"

        try:
            start_time = time.monotonic()
            response = requests.post(
                url=f"{base_url}{endpoint}",
                json=payload,
                stream=stream,
                headers=self.headers,
            )
            non_stream_post_end_time = time.monotonic()

            if response.status_code == 200:
                metrics_response = parse_strategy(
                    response,
                    start_time,
                    num_prefill_tokens,
                    non_stream_post_end_time,
                    model_name,
                )
            else:
                metrics_response = UserResponse(
                    status_code=response.status_code,
                    error_message=response.text,
                )
        except requests.exceptions.ConnectionError as e:
            metrics_response = UserResponse(
                status_code=503, error_message=f"Connection error: {e}"
            )
        except requests.exceptions.Timeout as e:
            metrics_response = UserResponse(
                status_code=408, error_message=f"Request timed out: {e}"
            )
        except requests.exceptions.RequestException as e:
            metrics_response = UserResponse(
                status_code=500,
                error_message=str(e),
            )
        finally:
            if response is not None:
                response.close()

        self.collect_metrics(metrics_response, endpoint)
        return metrics_response

    def parse_chat_response(
        self,
        response: Response,
        start_time: float,
        num_prefill_tokens: int,
        _: float,
        model_name: Optional[str] = None,
    ) -> UserResponse:
        """
        Parses a Gemini streaming response.

        Args:
            response (Response): The response object.
            start_time (float): The time when the request was started.
            num_prefill_tokens (int): The num of tokens in the prefill/prompt.
            _ (float): Placeholder for an unused var.
            model_name (Optional[str]): The model name.

        Returns:
            UserChatResponse: A response object with metrics and generated text.
        """
        generated_text = ""
        time_at_first_token = None

        for line in response.iter_lines():
            if line:
                try:
                    # Vertex AI returns JSON objects, not SSE
                    chunk = json.loads(line)

                    if time_at_first_token is None:
                        time_at_first_token = time.monotonic()

                    # Extract content from chunk
                    candidates = chunk.get("candidates", [])
                    if candidates:
                        content = candidates[0].get("content", {})
                        parts = content.get("parts", [])
                        for part in parts:
                            text = part.get("text", "")
                            if text:
                                generated_text += text
                except json.JSONDecodeError:
                    continue

        end_time = time.monotonic()

        # Estimate tokens if not provided
        tokens_received = self.environment.sampler.get_token_length(
            generated_text, add_special_tokens=False
        )

        return UserChatResponse(
            status_code=200,
            generated_text=generated_text,
            tokens_received=tokens_received,
            time_at_first_token=time_at_first_token,
            num_prefill_tokens=num_prefill_tokens,
            start_time=start_time,
            end_time=end_time,
        )

    def parse_palm_response(
        self,
        response: Response,
        start_time: float,
        num_prefill_tokens: int,
        end_time: float,
        model_name: Optional[str] = None,
    ) -> UserResponse:
        """
        Parses a PaLM/non-Gemini response.

        Args:
            response (Response): The response object.
            start_time (float): The time when the request was started.
            num_prefill_tokens (int): The num of tokens in the prefill/prompt.
            end_time (float): The time when the request was finished.
            model_name (Optional[str]): The model name.

        Returns:
            UserChatResponse: A response object with metrics and generated text.
        """
        data = response.json()

        # Extract response text
        predictions = data.get("predictions", [])
        response_text = predictions[0].get("content", "") if predictions else ""

        # Estimate tokens
        tokens_received = self.environment.sampler.get_token_length(
            response_text, add_special_tokens=False
        )

        return UserChatResponse(
            status_code=200,
            generated_text=response_text,
            tokens_received=tokens_received,
            time_at_first_token=end_time,
            num_prefill_tokens=num_prefill_tokens,
            start_time=start_time,
            end_time=end_time,
        )

    @staticmethod
    def parse_embedding_response(
        response: Response,
        start_time: float,
        _: Optional[int],
        end_time: float,
        model_name: Optional[str] = None,
    ) -> UserResponse:
        """
        Parses an embedding response.

        Args:
            response (Response): The response object.
            start_time (float): The time when the request was started.
            _ (Optional[int]): Placeholder for unused parameter.
            end_time (float): The time when the request was finished.
            model_name (Optional[str]): The model name.

        Returns:
            UserResponse: A response object with metrics.
        """
        data = response.json()

        # Extract embeddings - count tokens from all predictions
        predictions = data.get("predictions", [])
        num_embeddings = len(predictions)

        return UserResponse(
            status_code=200,
            start_time=start_time,
            end_time=end_time,
            time_at_first_token=end_time,
            num_prefill_tokens=num_embeddings,  # Use number of embeddings as a proxy
        )

    def _prepare_request_body(
        self, request: UserChatRequest, model_name: str
    ) -> Dict[str, Any]:
        """Prepare request body based on model type.

        Args:
            request: User request
            model_name: Model name

        Returns:
            Request body dict
        """
        is_gemini = "gemini" in model_name.lower()
        body: Dict[str, Any]

        if is_gemini:
            # Gemini models use different format
            contents = []

            if isinstance(request, UserImageChatRequest) and request.image_content:
                # Multimodal request
                parts: List[Dict[str, Any]] = [{"text": request.prompt}]
                for image in request.image_content:
                    if image.startswith("data:image/"):
                        # Extract base64 data from data URL for inline data
                        image_data = image.split(",", 1)[1]
                        parts.append(
                            {
                                "inlineData": {
                                    "mimeType": "image/jpeg",
                                    "data": image_data,
                                }
                            }
                        )
                    elif image.startswith(("http://", "https://", "gs://")):
                        # Use fileData for HTTP URLs and Cloud Storage URIs
                        parts.append(
                            {
                                "fileData": {
                                    "mimeType": "image/jpeg",
                                    "fileUri": image,
                                }
                            }
                        )
                    else:
                        raise ValueError(
                            f"Unsupported image format for GCP Vertex AI: {type(image)}"
                        )

                contents.append({"parts": parts})
            else:
                # Text-only request
                contents.append({"parts": [{"text": request.prompt}]})

            # Build request body
            body = {"contents": contents}

            # Add generation config
            generation_config = {}
            if request.max_tokens:
                generation_config["maxOutputTokens"] = request.max_tokens
            if request.additional_request_params.get("temperature") is not None:
                generation_config["temperature"] = request.additional_request_params[
                    "temperature"
                ]
            if request.additional_request_params.get("top_p") is not None:
                generation_config["topP"] = request.additional_request_params["top_p"]
            if request.additional_request_params.get("top_k") is not None:
                generation_config["topK"] = request.additional_request_params["top_k"]

            if generation_config:
                body["generationConfig"] = generation_config

        else:
            # PaLM and other models
            body = {
                "instances": [{"content": request.prompt}],
                "parameters": {},
            }

            # Add parameters
            if request.max_tokens:
                body["parameters"]["maxOutputTokens"] = request.max_tokens
            if request.additional_request_params.get("temperature") is not None:
                body["parameters"]["temperature"] = request.additional_request_params[
                    "temperature"
                ]
            if request.additional_request_params.get("top_p") is not None:
                body["parameters"]["topP"] = request.additional_request_params["top_p"]
            if request.additional_request_params.get("top_k") is not None:
                body["parameters"]["topK"] = request.additional_request_params["top_k"]

        # Add any additional params
        if request.additional_request_params:
            if is_gemini:
                # For Gemini, merge additional params at the top level
                for key, value in request.additional_request_params.items():
                    if key not in ["temperature", "top_p", "top_k", "stream"]:
                        body[key] = value
            else:
                # For other models, merge into parameters
                body["parameters"].update(request.additional_request_params)

        return body



import os
from typing import Any, Dict, Optional

from genai_bench.auth.auth_provider import AuthProvider


class OpenAIAuth(AuthProvider):
    """OpenAI Authentication Provider."""

    def __init__(self, api_key: Optional[str] = None):
        """Initialize OpenAI Auth Provider.

        Args:
            api_key (Optional[str]): OpenAI API key. If None, will try to get from
            OPENAI_API_KEY environment variable.

        Raises:
            ValueError: If no API key is provided or found in environment
        """
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        if not self.api_key or not self.api_key.strip():
            raise ValueError(
                "OpenAI API key must be provided or set in "
                "OPENAI_API_KEY environment variable"
            )

    def get_config(self) -> Dict[str, Any]:
        """Get OpenAI configuration.

        Returns:
            Dict[str, Any]: Empty configuration dictionary
            as OpenAI doesn't need additional config
        """
        return {}

    def get_credentials(self) -> str:
        """Get OpenAI API key.

        Returns:
            str: OpenAI API key
        """
        if not self.api_key:
            raise ValueError("OpenAI API key is not set")
        return self.api_key



"""Plot configuration system for flexible, user-defined plots."""

import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, field_validator

from genai_bench.logging import init_logger
from genai_bench.metrics.metrics import AggregatedMetrics
from genai_bench.time_units import TimeUnitConverter

logger = init_logger(__name__)


class YFieldSpec(BaseModel):
    """Specification for a Y-axis field in multi-line plots."""

    field: str = Field(..., description="Field path for Y-axis data")
    label: Optional[str] = Field(None, description="Custom label for this line")
    color: Optional[str] = Field(None, description="Custom color for this line")
    linestyle: Optional[str] = Field(
        None, description="Line style: '-', '--', '-.', ':'"
    )


class PlotSpec(BaseModel):
    """Specification for a single plot."""

    title: str = Field(..., description="Title of the plot")
    x_field: str = Field(..., description="Field path for X-axis data")
    y_field: Optional[str] = Field(
        None, description="Field path for Y-axis data (single line)"
    )
    y_fields: Optional[List[YFieldSpec]] = Field(
        None, description="Multiple Y-fields for multi-line plots"
    )
    x_label: Optional[str] = Field(None, description="Custom X-axis label")
    y_label: Optional[str] = Field(None, description="Custom Y-axis label")
    plot_type: str = Field(
        default="line", description="Type of plot: line, scatter, bar"
    )
    position: Tuple[int, int] = Field(..., description="Position in grid (row, col)")
    y_scale: Optional[str] = Field(
        None, description="Y-axis scale type: 'linear' (default) or 'log'"
    )

    @field_validator("plot_type")
    def validate_plot_type(cls, v):
        allowed_types = {"line", "scatter", "bar"}
        if v not in allowed_types:
            raise ValueError(f"plot_type must be one of {allowed_types}, got {v}")
        return v

    @field_validator("y_scale")
    def validate_y_scale(cls, v):
        if v is not None:
            allowed_scales = {"linear", "log"}
            if v not in allowed_scales:
                raise ValueError(f"y_scale must be one of {allowed_scales}, got {v}")
        return v

    @field_validator("y_fields")
    def validate_y_fields(cls, v, info):
        """Ensure either y_field or y_fields is specified, but not both."""
        y_field = info.data.get("y_field") if hasattr(info, "data") else None

        if y_field is not None and v is not None:
            raise ValueError(
                "Cannot specify both y_field and y_fields. Use one or the other."
            )

        if y_field is None and (v is None or len(v) == 0):
            raise ValueError(
                "Must specify either y_field or y_fields with at least one field."
            )

        return v

    def get_y_field_specs(self) -> List[YFieldSpec]:
        """Get Y-field specifications, converting single y_field if needed."""
        if self.y_fields is not None:
            return self.y_fields
        elif self.y_field is not None:
            return [YFieldSpec(field=self.y_field)]  # type: ignore[call-arg]
        else:
            raise ValueError("No Y-field specifications found")

    def is_multi_line(self) -> bool:
        """Check if this plot has multiple Y-fields."""
        return self.y_fields is not None and len(self.y_fields) > 1


class PlotLayout(BaseModel):
    """Layout configuration for the plot grid."""

    rows: int = Field(default=2, ge=1, le=5, description="Number of rows")
    cols: int = Field(default=4, ge=1, le=6, description="Number of columns")
    figsize: Optional[Tuple[int, int]] = Field(
        None, description="Figure size (width, height)"
    )


class PlotConfig(BaseModel):
    """Complete plot configuration."""

    layout: PlotLayout = Field(default_factory=lambda: PlotLayout())  # type: ignore[call-arg]
    plots: List[PlotSpec] = Field(..., description="List of plot specifications")

    @field_validator("plots")
    def validate_plot_positions(cls, v, info):
        if not hasattr(info, "data") or "layout" not in info.data:
            return v

        layout = info.data["layout"]
        max_row, max_col = layout.rows - 1, layout.cols - 1

        positions = set()
        for plot in v:
            row, col = plot.position
            if row > max_row or col > max_col:
                raise ValueError(
                    f"Plot position ({row}, {col}) exceeds layout bounds "
                    f"({max_row}, {max_col})"
                )
            if plot.position in positions:
                raise ValueError(f"Duplicate plot position: {plot.position}")
            positions.add(plot.position)

        return v


class PlotConfigManager:
    """Manager for plot configurations."""

    # Built-in presets
    PRESETS = {
        "2x4_default": {
            "layout": {"rows": 2, "cols": 4, "figsize": [32, 12]},
            "plots": [
                {
                    "title": "Output Inference Speed per Request vs "
                    "Output Throughput of Server",
                    "x_field": "mean_output_throughput_tokens_per_s",
                    "y_field": "stats.output_inference_speed.mean",
                    "x_label": "Output Throughput of Server (tokens/s)",
                    "y_label": "Output Inference Speed per Request (tokens/s)",
                    "plot_type": "line",
                    "position": [0, 0],
                },
                {
                    "title": "TTFT vs Output Throughput of Server",
                    "x_field": "mean_output_throughput_tokens_per_s",
                    "y_field": "stats.ttft.mean",
                    "x_label": "Output Throughput of Server (tokens/s)",
                    "y_label": "TTFT (s)",
                    "plot_type": "line",
                    "position": [0, 1],
                },
                {
                    "title": "Mean E2E Latency per Request vs RPS",
                    "x_field": "requests_per_second",
                    "y_field": "stats.e2e_latency.mean",
                    "x_label": "RPS (req/s)",
                    "y_label": "Mean E2E Latency per Request (s)",
                    "plot_type": "line",
                    "position": [0, 2],
                },
                {
                    "title": "Error Rates by HTTP Status vs Concurrency",
                    "x_field": "num_concurrency",
                    "y_field": "error_rate",
                    "x_label": "Concurrency",
                    "y_label": "Error Rate",
                    "plot_type": "bar",
                    "position": [0, 3],
                },
                {
                    "title": "Output Inference Speed per Request vs "
                    "Total Throughput (Input + Output) of Server",
                    "x_field": "mean_total_tokens_throughput_tokens_per_s",
                    "y_field": "stats.output_inference_speed.mean",
                    "x_label": "Total Throughput (Input + Output) of Server (tokens/s)",
                    "y_label": "Output Inference Speed per Request (tokens/s)",
                    "plot_type": "line",
                    "position": [1, 0],
                },
                {
                    "title": "TTFT vs Total Throughput (Input + Output) of Server",
                    "x_field": "mean_total_tokens_throughput_tokens_per_s",
                    "y_field": "stats.ttft.mean",
                    "x_label": "Total Throughput (Input + Output) of Server (tokens/s)",
                    "y_label": "TTFT (s)",
                    "plot_type": "line",
                    "position": [1, 1],
                },
                {
                    "title": "P90 E2E Latency per Request vs RPS",
                    "x_field": "requests_per_second",
                    "y_field": "stats.e2e_latency.p90",
                    "x_label": "RPS (req/s)",
                    "y_label": "P90 E2E Latency per Request (s)",
                    "plot_type": "line",
                    "position": [1, 2],
                },
                {
                    "title": "P99 E2E Latency per Request vs RPS",
                    "x_field": "requests_per_second",
                    "y_field": "stats.e2e_latency.p99",
                    "x_label": "RPS (req/s)",
                    "y_label": "P99 E2E Latency per Request (s)",
                    "plot_type": "line",
                    "position": [1, 3],
                },
            ],
        },
        "simple_2x2": {
            "layout": {"rows": 2, "cols": 2, "figsize": [16, 12]},
            "plots": [
                {
                    "title": "Throughput vs Latency",
                    "x_field": "mean_output_throughput_tokens_per_s",
                    "y_field": "stats.e2e_latency.mean",
                    "plot_type": "line",
                    "position": [0, 0],
                },
                {
                    "title": "RPS vs Error Rate",
                    "x_field": "requests_per_second",
                    "y_field": "error_rate",
                    "plot_type": "scatter",
                    "position": [0, 1],
                },
                {
                    "title": "Concurrency vs TTFT",
                    "x_field": "num_concurrency",
                    "y_field": "stats.ttft.mean",
                    "plot_type": "line",
                    "position": [1, 0],
                },
                {
                    "title": "Throughput Distribution",
                    "x_field": "num_concurrency",
                    "y_field": "mean_total_tokens_throughput_tokens_per_s",
                    "plot_type": "line",
                    "position": [1, 1],
                },
            ],
        },
        "multi_line_latency": {
            "layout": {"rows": 2, "cols": 2, "figsize": [16, 12]},
            "plots": [
                {
                    "title": "Latency Percentiles Comparison",
                    "x_field": "requests_per_second",
                    "y_fields": [
                        {
                            "field": "stats.e2e_latency.mean",
                            "label": "Mean Latency",
                            "color": "blue",
                            "linestyle": "-",
                        },
                        {
                            "field": "stats.e2e_latency.p90",
                            "label": "P90 Latency",
                            "color": "orange",
                            "linestyle": "--",
                        },
                        {
                            "field": "stats.e2e_latency.p99",
                            "label": "P99 Latency",
                            "color": "red",
                            "linestyle": "-.",
                        },
                    ],
                    "x_label": "Requests Per Second",
                    "y_label": "Latency (s)",
                    "plot_type": "line",
                    "position": [0, 0],
                },
                {
                    "title": "TTFT Performance Analysis",
                    "x_field": "mean_output_throughput_tokens_per_s",
                    "y_fields": [
                        {
                            "field": "stats.ttft.mean",
                            "label": "Mean TTFT (s)",
                            "color": "green",
                        },
                        {
                            "field": "stats.ttft.p95",
                            "label": "P95 TTFT (s)",
                            "color": "purple",
                        },
                    ],
                    "x_label": "Output Throughput (tokens/s)",
                    "y_label": "Time to First Token (s) (log scale)",
                    "plot_type": "line",
                    "position": [0, 1],
                    "y_scale": "log",
                },
                {
                    "title": "Throughput Components",
                    "x_field": "num_concurrency",
                    "y_fields": [
                        {
                            "field": "mean_output_throughput_tokens_per_s",
                            "label": "Output Throughput",
                        },
                        {
                            "field": "mean_total_tokens_throughput_tokens_per_s",
                            "label": "Total Throughput",
                        },
                    ],
                    "x_label": "Concurrency Level",
                    "y_label": "Throughput (tokens/s)",
                    "plot_type": "line",
                    "position": [1, 0],
                },
                {
                    "title": "Request Success Rate",
                    "x_field": "num_concurrency",
                    "y_field": "error_rate",
                    "x_label": "Concurrency Level",
                    "y_label": "Error Rate",
                    "plot_type": "bar",
                    "position": [1, 1],
                },
            ],
        },
        "single_scenario_analysis": {
            "layout": {"rows": 2, "cols": 2, "figsize": [16, 12]},
            "plots": [
                {
                    "title": "Latency Percentiles vs RPS",
                    "x_field": "requests_per_second",
                    "y_fields": [
                        {
                            "field": "stats.e2e_latency.mean",
                            "label": "Mean",
                            "color": "blue",
                            "linestyle": "-",
                        },
                        {
                            "field": "stats.e2e_latency.p90",
                            "label": "P90",
                            "color": "orange",
                            "linestyle": "--",
                        },
                        {
                            "field": "stats.e2e_latency.p99",
                            "label": "P99",
                            "color": "red",
                            "linestyle": "-.",
                        },
                    ],
                    "x_label": "Requests Per Second",
                    "y_label": "E2E Latency (s)",
                    "plot_type": "line",
                    "position": [0, 0],
                },
                {
                    "title": "TTFT Analysis",
                    "x_field": "mean_output_throughput_tokens_per_s",
                    "y_fields": [
                        {
                            "field": "stats.ttft.mean",
                            "label": "Mean TTFT (s)",
                            "color": "green",
                        },
                        {
                            "field": "stats.ttft.p95",
                            "label": "P95 TTFT (s)",
                            "color": "purple",
                        },
                    ],
                    "x_label": "Output Throughput (tokens/s)",
                    "y_label": "TTFT (s)",
                    "plot_type": "line",
                    "position": [0, 1],
                },
                {
                    "title": "Throughput vs Concurrency",
                    "x_field": "num_concurrency",
                    "y_field": "mean_output_throughput_tokens_per_s",
                    "x_label": "Concurrency",
                    "y_label": "Output Throughput (tokens/s)",
                    "plot_type": "line",
                    "position": [1, 0],
                },
                {
                    "title": "Error Rate vs Concurrency",
                    "x_field": "num_concurrency",
                    "y_field": "error_rate",
                    "x_label": "Concurrency",
                    "y_label": "Error Rate",
                    "plot_type": "bar",
                    "position": [1, 1],
                },
            ],
        },
    }

    @classmethod
    def apply_time_unit_conversion(
        cls, config_data: Dict[str, Any], metrics_time_unit: str = "s"
    ) -> Dict[str, Any]:
        """
        Apply time unit conversion to plot configuration labels.

        Args:
            config_data: Plot configuration dictionary
            metrics_time_unit: Target time unit ('s' or 'ms')

        Returns:
            Modified configuration with updated labels
        """
        if metrics_time_unit not in ["s", "ms"]:
            return config_data

        # Create a copy to avoid modifying the original
        config_copy = config_data.copy()

        # Convert labels in each plot specification
        for plot_copy in config_copy.get("plots", []):
            # Convert title
            if "title" in plot_copy:
                original_title = plot_copy["title"]
                plot_copy["title"] = TimeUnitConverter.get_unit_label(
                    original_title, metrics_time_unit
                )

            # Convert y_label
            if "y_label" in plot_copy:
                original_y_label = plot_copy["y_label"]
                plot_copy["y_label"] = TimeUnitConverter.get_unit_label(
                    original_y_label, metrics_time_unit
                )

            # Convert x_label
            if "x_label" in plot_copy:
                original_x_label = plot_copy["x_label"]
                plot_copy["x_label"] = TimeUnitConverter.get_unit_label(
                    original_x_label, metrics_time_unit
                )

            # Convert y_fields labels
            if "y_fields" in plot_copy:
                for y_field in plot_copy["y_fields"]:
                    if "label" in y_field:
                        original_label = y_field["label"]
                        y_field["label"] = TimeUnitConverter.get_unit_label(
                            original_label, metrics_time_unit
                        )

        return config_copy

    @classmethod
    def load_config(
        cls, config_source: Union[str, Dict, None] = None, metrics_time_unit: str = "s"
    ) -> PlotConfig:
        """Load plot configuration from various sources."""
        if config_source is None:
            # Use default 2x4 preset
            return cls.load_preset("2x4_default", metrics_time_unit)

        if isinstance(config_source, str):
            if config_source in cls.PRESETS:
                # Load preset
                return cls.load_preset(config_source, metrics_time_unit)
            else:
                # Load from file
                return cls.load_from_file(config_source, metrics_time_unit)

        if isinstance(config_source, dict):
            # Load from dictionary
            converted_data = cls.apply_time_unit_conversion(
                config_source, metrics_time_unit
            )
            return PlotConfig(**converted_data)

        raise ValueError(f"Invalid config source type: {type(config_source)}")

    @classmethod
    def load_preset(cls, preset_name: str, metrics_time_unit: str = "s") -> PlotConfig:
        """Load a preset plot configuration."""
        if preset_name not in cls.PRESETS:
            raise ValueError(
                f"Unknown preset: {preset_name}. Available: {list(cls.PRESETS.keys())}"
            )

        preset_data = cls.PRESETS[preset_name]
        return cls.load_config(preset_data, metrics_time_unit)

    @classmethod
    def load_from_file(cls, file_path: str, metrics_time_unit: str = "s") -> PlotConfig:
        """Load configuration from JSON file."""
        path = Path(file_path)
        if not path.exists():
            raise FileNotFoundError(f"Plot config file not found: {file_path}")

        try:
            with open(path, "r") as f:
                config_data = json.load(f)
            converted_data = cls.apply_time_unit_conversion(
                config_data, metrics_time_unit
            )
            return PlotConfig(**converted_data)
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in config file {file_path}: {e}") from e
        except Exception as e:
            raise ValueError(f"Error loading config from {file_path}: {e}") from e

    @classmethod
    def save_config(cls, config: PlotConfig, file_path: str) -> None:
        """Save configuration to JSON file."""
        path = Path(file_path)
        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "w") as f:
            json.dump(config.model_dump(), f, indent=2)

    @classmethod
    def get_available_fields(cls) -> Dict[str, str]:
        """Get all available fields from AggregatedMetrics."""
        fields = {}

        # Direct fields
        for field_name, field_info in AggregatedMetrics.model_fields.items():
            if field_name != "stats":
                fields[field_name] = field_info.description or field_name

        # Stats fields with statistical measures
        stats_measures = [
            "min",
            "max",
            "mean",
            "stddev",
            "sum",
            "p25",
            "p50",
            "p75",
            "p90",
            "p95",
            "p99",
        ]
        stats_fields = [
            "ttft",
            "tpot",
            "e2e_latency",
            "output_latency",
            "output_inference_speed",
            "num_input_tokens",
            "num_output_tokens",
            "total_tokens",
            "input_throughput",
            "output_throughput",
        ]

        for field in stats_fields:
            for measure in stats_measures:
                field_path = f"stats.{field}.{measure}"
                fields[field_path] = f"{field} {measure}"

        return fields

    @classmethod
    def get_fields_from_data(
        cls, metrics: AggregatedMetrics
    ) -> Dict[str, Tuple[Any, str]]:
        """
        Extract available fields with actual values from real experiment data.

        Returns:
            Dict mapping field_path -> (value, type_name)
        """
        available_fields = {}

        # Direct fields from AggregatedMetrics
        for field_name in AggregatedMetrics.model_fields:
            if field_name == "stats":
                continue

            try:
                value = getattr(metrics, field_name)
                if value is not None:
                    field_type = type(value).__name__
                    available_fields[field_name] = (value, field_type)
            except AttributeError:
                continue

        # Stats fields - check which ones have actual data
        if hasattr(metrics, "stats") and metrics.stats is not None:
            stats = metrics.stats

            # Get available stat fields from the stats object
            for stat_field_name in stats.model_fields:
                try:
                    stat_field = getattr(stats, stat_field_name)
                    if stat_field is not None:
                        # Check each statistical measure
                        for measure_name in stat_field.model_fields:
                            try:
                                measure_value = getattr(stat_field, measure_name)
                                if measure_value is not None:
                                    field_path = (
                                        f"stats.{stat_field_name}.{measure_name}"
                                    )
                                    field_type = type(measure_value).__name__
                                    available_fields[field_path] = (
                                        measure_value,
                                        field_type,
                                    )
                            except AttributeError:
                                continue
                except AttributeError:
                    continue

        return available_fields

    @classmethod
    def validate_field_path(
        cls, field_path: str, sample_metrics: AggregatedMetrics
    ) -> bool:
        """Validate that a field path exists in AggregatedMetrics."""
        try:
            value = cls.get_field_value(sample_metrics, field_path)
            return value is not None
        except Exception:
            return False

    @classmethod
    def get_field_value(cls, metrics: AggregatedMetrics, field_path: str) -> Any:
        """Get value from metrics using field path (e.g., 'stats.ttft.mean')."""
        parts = field_path.split(".")
        value = metrics

        for part in parts:
            if hasattr(value, part):
                value = getattr(value, part)
            else:
                raise AttributeError(f"Field path '{field_path}' not found")

        return value



from locust.env import Environment

import time
from datetime import datetime
from pathlib import Path
from typing import Optional, Tuple

from genai_bench.logging import init_logger

logger = init_logger(__name__)


def manage_run_time(
    max_time_per_run: int,
    max_requests_per_run: int,
    environment: Environment,
) -> int:
    """
    Manages the run time of the benchmarking process by tracking elapsed time
    and ensuring enough requests are completed before the test ends. The
    function will exit when one of the two conditions is met:
    1. The maximum allowed run time (`max_time_per_run`) is reached.
    2. The total number of requests exceeds the maximum requests per run
       (`max_requests_per_run`).

    Args:
        max_time_per_run (int): The maximum allowed run time in seconds.
        max_requests_per_run (int): The maximum number of requests per
            run.
        environment: The environment object with runner stats.

    Returns:
        int: The actual run time in seconds.
    """

    total_run_time = 0

    while total_run_time < max_time_per_run:
        time.sleep(1)
        total_run_time += 1

        assert environment.runner is not None, "environment.runner should not be None"
        total_completed_requests = environment.runner.stats.total.num_requests

        if total_completed_requests >= max_requests_per_run:
            logger.info(
                f"⏩ Exit the run as {total_completed_requests} requests have "
                "been completed."
            )
            break

    return int(total_run_time)


def get_experiment_path(
    experiment_folder_name: Optional[str],
    experiment_base_dir: Optional[str],
    api_backend: str,
    server_engine: Optional[str],
    server_version: Optional[str],
    task: str,
    model: str,
) -> Path:
    """
    Generate experiment path based on provided options and configuration.

    Args:
        experiment_folder_name: Optional custom folder name
        experiment_base_dir: Optional base directory for experiments
            (relative or absolute)
        api_backend: API backend name
        server_engine: Optional server engine name
        server_version: Optional server version
        task: Task name
        model: Model name

    Returns:
        Path: Full path to experiment directory
    """
    # Generate default name if not provided
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    default_name = (
        f"{api_backend}_"
        f"{server_engine + '_' if server_engine else ''}"
        f"{server_version + '_' if server_version else ''}"
        f"{task}_{model}_{timestamp}"
    )

    # Use provided name or default
    folder_name = experiment_folder_name or default_name

    # Determine full path
    if experiment_base_dir:
        # Convert to absolute path if relative
        base_dir = Path(experiment_base_dir).resolve()
        base_dir.mkdir(parents=True, exist_ok=True)
        experiment_path = base_dir / folder_name
    else:
        experiment_path = Path(folder_name)

    if experiment_path.exists():
        logger.warning(
            f"‼️ The folder {experiment_path} already exists. Data might be overridden."
        )

    experiment_path.mkdir(parents=True, exist_ok=True)
    return experiment_path


def get_run_params(iteration_type: str, iteration_value: int) -> Tuple[str, int, int]:
    """
    Returns appropriate header, batch_size, and num_concurrency based on iteration_type
    and iteration_value.
    """
    if iteration_type == "batch_size":
        return "Batch Size", iteration_value, 1
    return "Concurrency", 1, iteration_value



"""Interface definitions for data store operations."""

from abc import ABC, abstractmethod
from typing import Optional

from genai_bench.storage.oci_object_storage.object_uri import ObjectURI


class DataStore(ABC):
    """Interface for data store operations."""

    @abstractmethod
    def download(
        self, source: ObjectURI, target: str, retries: Optional[int] = 3
    ) -> None:
        """Download an object from its source path to the target path.

        Args:
            source: Source object URI
            target: Target local path
            retries: Number of retry attempts

        Raises:
            ServiceError: If object retrieval fails
            OSError: If file operations fail
        """
        pass

    @abstractmethod
    def upload(
        self, source: str, target: ObjectURI, retries: Optional[int] = 3
    ) -> None:
        """Upload an object from its source path to the target path.

        Args:
            source: Source local path
            target: Target object URI
            retries: Number of retry attempts

        Raises:
            ServiceError: If upload fails
            OSError: If file operations fail
        """
        pass



from locust import task

import json
import time
from typing import Any, Callable, List, Optional

from oci.generative_ai_inference import GenerativeAiInferenceClient
from oci.generative_ai_inference.models import (
    ChatDetails,
    ChatResult,
    CohereChatRequest,
    DedicatedServingMode,
    EmbedTextDetails,
    EmbedTextResult,
    OnDemandServingMode,
    RerankTextDetails,
)

from genai_bench.auth.model_auth_provider import ModelAuthProvider
from genai_bench.logging import init_logger
from genai_bench.protocol import (
    UserChatRequest,
    UserChatResponse,
    UserEmbeddingRequest,
    UserImageEmbeddingRequest,
    UserRequest,
    UserReRankRequest,
    UserResponse,
)
from genai_bench.user.base_user import BaseUser

logger = init_logger(__name__)


class OCICohereUser(BaseUser):
    """User class for Cohere model API with OCI authentication."""

    BACKEND_NAME = "oci-cohere"
    supported_tasks = {
        "text-to-text": "chat",
        "text-to-rerank": "rerank",
        "text-to-embeddings": "embeddings",
        "image-to-embeddings": "embeddings",
    }
    host: Optional[str] = None
    auth_provider: Optional[ModelAuthProvider] = None

    def on_start(self):
        """Initialize OCI client on start."""
        super().on_start()
        if not self.auth_provider:
            raise ValueError("Auth is required for OCICohereUser")
        self.client = GenerativeAiInferenceClient(
            config=self.auth_provider.get_config(),
            signer=self.auth_provider.get_credentials(),
            service_endpoint=self.host,
        )
        logger.debug("Generative AI Inference Client initialized.")

    def send_request(
        self,
        make_request: Callable,
        endpoint: str,
        payload: Any,
        parse_strategy: Any,
        num_prefill_tokens: Optional[int] = None,
    ) -> UserResponse:
        logger.debug(f"Sending request with payload: {payload}")
        try:
            start_time = time.monotonic()
            response = make_request()
            non_stream_post_end_time = time.monotonic()

            if response.status == 200:
                metrics_response = parse_strategy(
                    payload,
                    response,
                    start_time,
                    num_prefill_tokens,
                    non_stream_post_end_time,
                )
            else:
                logger.warning(
                    f"Received error status-code: {response.status} "
                    f"RequestId: {response.request_id}",
                    f"Response: {response.response}",
                )
                metrics_response = UserResponse(
                    status_code=response.status,
                    error_message="Request Failed",
                )
            self.collect_metrics(metrics_response, endpoint)
            return metrics_response
        except Exception as e:
            return UserResponse(
                status_code=500,
                error_message=str(e),
                num_prefill_tokens=num_prefill_tokens or 0,
            )

    @task
    def chat(self):
        """Send a chat completion request using Cohere format."""
        user_request = self.sample()

        if not isinstance(user_request, UserChatRequest):
            raise AttributeError(
                f"Expected UserChatRequest for OCICohereUser.chat, got "
                f"{type(user_request)}"
            )

        compartment_id = self.get_compartment_id(user_request)
        serving_mode = self.get_serving_mode(user_request)

        # Construct chat request
        chat_request = CohereChatRequest(
            api_format="COHERE",
            message=user_request.prompt,
            max_tokens=user_request.max_tokens,
            is_stream=True,
            temperature=user_request.additional_request_params.get("temperature", 0.1),
            top_p=user_request.additional_request_params.get("topP", 0),
            top_k=user_request.additional_request_params.get("topK", 0.75),
            frequency_penalty=user_request.additional_request_params.get(
                "frequencyPenalty", 0
            ),
            presence_penalty=user_request.additional_request_params.get(
                "presencePenalty", 0
            ),
        )

        # Add chat history and documents if provided
        if "chatHistory" in user_request.additional_request_params:
            chat_request.chat_history = user_request.additional_request_params[
                "chatHistory"
            ]
            logger.info(
                f"Chat history provided with {len(chat_request.chat_history)} items."
            )

        if "documents" in user_request.additional_request_params:
            chat_request.documents = user_request.additional_request_params["documents"]
            logger.info(f"Documents provided with {len(chat_request.documents)} items.")

        # Define payload with compartment ID and serving mode
        chat_detail = ChatDetails(
            compartment_id=compartment_id,
            serving_mode=serving_mode,
            chat_request=chat_request,
        )
        return self.send_request(
            make_request=lambda: self.client.chat(chat_detail),
            endpoint="chat",
            payload=chat_detail,
            parse_strategy=self.parse_chat_response,
            num_prefill_tokens=user_request.num_prefill_tokens,
        )

    @task
    def embeddings(self):
        """Send an embedding request using Cohere format."""
        user_request = self.sample()

        if not isinstance(user_request, UserEmbeddingRequest):
            raise AttributeError(
                f"user_request should be of type UserEmbeddingRequest for "
                f"OCICohereUser.embeddings, got {type(user_request)}"
            )

        if user_request.documents and not user_request.num_prefill_tokens:
            logger.warning(
                "Number of prefill tokens is missing or 0. Please double check."
            )

        # Retrieve compartment ID and serving mode
        compartment_id = self.get_compartment_id(user_request)
        serving_mode = self.get_serving_mode(user_request)
        input_type = self.get_embedding_input_type(user_request)
        inputs = self.get_inputs(user_request)

        embed_text_detail = EmbedTextDetails(
            compartment_id=compartment_id,
            serving_mode=serving_mode,
            inputs=inputs,
            input_type=input_type,
            truncate=user_request.additional_request_params.get("truncate", "NONE"),
        )

        response = self.send_request(
            make_request=lambda: self.client.embed_text(embed_text_detail),
            endpoint="embedText",
            payload=embed_text_detail,
            parse_strategy=self.parse_embedding_response,
            num_prefill_tokens=user_request.num_prefill_tokens,
        )

        logger.debug(f"Response received {response}")

    @task
    def rerank(self):
        """Send an rerank request using Cohere format."""
        user_request = self.sample()

        if not isinstance(user_request, UserReRankRequest):
            raise AttributeError(
                f"user_request should be of type UserReRankRequest for "
                f"OCICohereUser.rerank, got {type(user_request)}"
            )

        if user_request.documents and not user_request.num_prefill_tokens:
            logger.error(
                "Number of prefill tokens is missing or 0. Please double check."
            )

        # Retrieve compartment ID and serving mode
        compartment_id = self.get_compartment_id(user_request)
        top_n = self.get_top_n(user_request)
        serving_mode = self.get_serving_mode(user_request)
        documents = self.get_documents(user_request)
        query = self.get_query(user_request)

        # TODO: Re-rank3.5 API Changes for OCI GenAI are in progress
        rerank_text_details = RerankTextDetails(
            compartment_id=compartment_id,
            serving_mode=serving_mode,
            documents=documents,
            input=query,
            top_n=top_n,
        )

        response = self.send_request(
            make_request=lambda: self.client.rerank_text(rerank_text_details),
            endpoint="rerankText",
            payload=rerank_text_details,
            parse_strategy=self.parse_rerank_response,
            num_prefill_tokens=user_request.num_prefill_tokens,
        )

        logger.debug(f"Response received {response}")

    def get_compartment_id(self, user_request: UserRequest):
        compartment_id = user_request.additional_request_params.get("compartmentId")
        if not compartment_id:
            raise ValueError("compartmentId missing in additional request params")
        return compartment_id

    def get_top_n(self, user_request: UserRequest):
        return user_request.additional_request_params.get("topN")

    def get_embedding_input_type(self, user_request) -> str:
        input_type = "SEARCH_DOCUMENT"
        if isinstance(user_request, UserImageEmbeddingRequest):
            input_type = "IMAGE"
        return input_type

    def get_documents(self, re_rank_request: UserReRankRequest) -> List[Any]:
        return re_rank_request.documents

    def get_query(self, re_rank_request: UserReRankRequest) -> str:
        return re_rank_request.query

    def get_inputs(self, user_request) -> List[Any]:
        if isinstance(user_request, UserImageEmbeddingRequest):
            num_sampled_images = len(user_request.image_content)
            if num_sampled_images > 1:
                raise ValueError(
                    f"OCI-Cohere Image embedding supports only 1 "
                    f"image but, the value provided in traffic"
                    f"scenario is requesting {num_sampled_images}"
                )
            return user_request.image_content
        return user_request.documents

    def get_serving_mode(self, user_request: UserRequest) -> Any:
        params = user_request.additional_request_params
        model_id = user_request.model
        serving_type = params.get("servingType", "ON_DEMAND")
        if serving_type == "DEDICATED":
            endpoint_id = params.get("endpointId")
            if not endpoint_id:
                raise ValueError(
                    "endpointId must be provided for DEDICATED servingType"
                )
            logger.debug(
                f"Using DedicatedServingMode {serving_type} with "
                f"endpoint ID: {endpoint_id}"
            )
            return DedicatedServingMode(endpoint_id=endpoint_id)
        else:
            logger.debug(
                f"Using OnDemandServingMode {serving_type} with model ID: {model_id}"
            )
            return OnDemandServingMode(model_id=model_id)

    def parse_chat_response(
        self,
        request: ChatDetails,
        response: ChatResult,
        start_time: float,
        num_prefill_tokens: int,
        _: float,
    ) -> UserResponse:
        """
        Parses the streaming response from the Cohere API in OCI format.

        Args:
            request (ChatDetails): OCICohere Chat request.
            response (ChatResult): The streaming response from the Cohere API.
            start_time (float): Timestamp of request initiation.
            num_prefill_tokens (int): Number of tokens in the prompt.
            _ (float): Placeholder for unused variable.

        Returns:
            UserResponse: Parsed response in the UserResponse format.
        """
        generated_text = ""
        tokens_received = 0
        time_at_first_token: Optional[float] = None
        finish_reason = None
        previous_data = None

        # Iterate over each event in the streaming response
        for event in response.data.events():
            # Raw event data from the stream
            event_data = event.data.strip()

            # Parse the event data as JSON
            try:
                parsed_data = json.loads(event_data)
                finish_reason = parsed_data.get("finishReason", None)
                if not finish_reason:
                    # Extract text content if present
                    text_segment = parsed_data.get("text", "")
                    if text_segment:
                        # Capture the time at the first token
                        if not time_at_first_token:
                            time_at_first_token = time.monotonic()
                        generated_text += text_segment
                        tokens_received += 1  # each event contains one token
                        logger.debug(f"number of tokens received: {tokens_received}")
                    # Track the previous data for debugging purposes
                    previous_data = parsed_data
                else:
                    # we have reached the end
                    logger.debug(
                        f"We have reached the end of the response "
                        f"with finish reason: {finish_reason}"
                    )
                    break
            except json.JSONDecodeError:
                logger.warning(
                    f"Error decoding JSON from event data: {event_data}, "
                    f"previous data: {previous_data}, "
                    f"finish reason: {finish_reason}"
                )
                continue

        # End timing for response
        end_time = time.monotonic()
        logger.debug(
            f"Generated text: {generated_text} \n"
            f"Time at first token: {time_at_first_token} \n"
            f"Finish reason: {finish_reason}\n"
            f"Completion Tokens: {tokens_received}\n"
            f"Start Time: {start_time}\n"
            f"End Time: {end_time}"
        )
        # Log if token count was not captured accurately
        if not tokens_received:
            tokens_received = len(generated_text.split())

        return UserChatResponse(
            status_code=200,
            generated_text=generated_text,
            tokens_received=tokens_received,
            time_at_first_token=time_at_first_token,
            num_prefill_tokens=num_prefill_tokens,
            start_time=start_time,
            end_time=end_time,
        )

    def parse_embedding_response(
        self,
        request: EmbedTextDetails,
        _: EmbedTextResult,
        start_time: float,
        num_prefill_tokens: int,
        end_time: float,
    ) -> UserResponse:
        """
        Parses a non-streaming response.

        Args:
            request (EmbedTextDetails): The request object.
            _ (EmbedTextResult): The response object.
            start_time (float): The time when the request was started.
            num_prefill_tokens (int): Number of tokens in the prefill/prompt.
            end_time(float): The time when the request was finished.

        Returns:
            UserResponse: A response object with metrics.
        """
        if num_prefill_tokens is None:
            num_prefill_tokens = len(request.inputs)

        return UserResponse(
            status_code=200,
            num_prefill_tokens=num_prefill_tokens,
            start_time=start_time,
            end_time=end_time,
            time_at_first_token=end_time,
        )

    def parse_rerank_response(
        self,
        request: RerankTextDetails,
        _: RerankTextDetails,
        start_time: float,
        num_prefill_tokens: int,
        end_time: float,
    ) -> UserResponse:
        """
        Parses a non-streaming response.

        Args:
            request (RerankTextDetails): The request object.
            _ (RerankTextDetails): The response object.
            start_time (float): The time when the request was started.
            num_prefill_tokens (int): Number of tokens in the prefill/prompt.
            end_time(float): The time when the request was finished.

        Returns:
            UserResponse: A response object with metrics.
        """

        return UserResponse(
            status_code=200,
            num_prefill_tokens=num_prefill_tokens,
            start_time=start_time,
            end_time=end_time,
            time_at_first_token=end_time,
        )



"""OCI Object URI data class."""

import os
from typing import Optional

from pydantic import BaseModel, Field


class ObjectURI(BaseModel):
    """Represents an OCI Storage URI.

    Format: oci://n/{namespace}/b/{bucket}/o/{object_path}
    """

    namespace: str = Field(description="OCI namespace")
    bucket_name: str = Field(description="Bucket name")
    object_name: Optional[str] = Field(None, description="Object name/path")
    region: Optional[str] = Field(None, description="OCI region")
    prefix: Optional[str] = Field(None, description="Object prefix for listing")

    @classmethod
    def from_uri(cls, uri: str) -> "ObjectURI":
        """Create ObjectURI from string URI.

        Args:
            uri: OCI storage URI string

        Returns:
            ObjectURI object

        Raises:
            ValueError: If URI format is invalid
        """
        if not uri.startswith("oci://"):
            raise ValueError(f"Invalid OCI storage URI format: {uri}")

        # Remove oci:// prefix
        path = uri[6:]

        try:
            # Parse namespace
            if not path.startswith("n/"):
                raise ValueError("Missing namespace (n/) in URI")
            path = path[2:]
            namespace, path = path.split("/", 1)

            # Parse bucket
            if not path.startswith("b/"):
                raise ValueError("Missing bucket (b/) in URI")
            path = path[2:]
            bucket, path = path.split("/", 1)

            # Parse object path (optional)
            object_name = None
            prefix = None
            if path.startswith("o/"):
                path = path[2:]
                if path:
                    object_name = path
                    prefix = os.path.dirname(object_name)
                    if prefix:
                        prefix = prefix + "/"

            return cls(
                namespace=namespace,
                bucket_name=bucket,
                object_name=object_name,
                region=None,  # Region is not part of the URI format
                prefix=prefix,
            )

        except ValueError as e:
            raise ValueError(f"Invalid OCI storage URI format: {uri}") from e

    def __str__(self) -> str:
        """Get string representation of URI.

        Returns:
            URI string in format oci://n/{namespace}/b/{bucket}/o/{object_name}
        """
        base = f"oci://n/{self.namespace}/b/{self.bucket_name}"
        if self.object_name:
            return f"{base}/o/{self.object_name}"
        return base



"""Unified factory for creating model and storage authentication providers."""

import oci.config

# Model auth providers
from genai_bench.auth.aws.bedrock_auth import AWSBedrockAuth

# Storage auth providers
from genai_bench.auth.aws.s3_auth import AWSS3Auth
from genai_bench.auth.azure.blob_auth import AzureBlobAuth
from genai_bench.auth.azure.openai_auth import AzureOpenAIAuth

# Import existing auth providers
from genai_bench.auth.factory import AuthFactory
from genai_bench.auth.gcp.gcs_auth import GCPStorageAuth
from genai_bench.auth.gcp.vertex_auth import GCPVertexAuth
from genai_bench.auth.github.github_auth import GitHubAuth
from genai_bench.auth.model_auth_provider import ModelAuthProvider
from genai_bench.auth.oci.model_auth_adapter import OCIModelAuthAdapter
from genai_bench.auth.oci.storage_auth_adapter import OCIStorageAuthAdapter
from genai_bench.auth.openai.auth import OpenAIAuth
from genai_bench.auth.openai.model_auth_adapter import OpenAIModelAuthAdapter
from genai_bench.auth.storage_auth_provider import StorageAuthProvider
from genai_bench.auth.together.auth import TogetherAuth
from genai_bench.auth.together.model_auth_adapter import TogetherModelAuthAdapter


class UnifiedAuthFactory:
    """Factory for creating model and storage authentication providers."""

    @staticmethod
    def create_model_auth(provider: str, **kwargs) -> ModelAuthProvider:
        """Create a model endpoint authentication provider.

        Args:
            provider: Provider type ('openai', 'oci', 'aws-bedrock',
                'azure-openai', 'gcp-vertex', 'together')
            **kwargs: Provider-specific arguments

        Returns:
            ModelAuthProvider instance

        Raises:
            ValueError: If provider is not supported
        """
        if provider == "openai":
            api_key = kwargs.get("api_key")
            openai_auth = OpenAIAuth(api_key=api_key)
            return OpenAIModelAuthAdapter(openai_auth)

        elif provider == "oci":
            # Use existing OCI auth factory
            oci_auth = AuthFactory.create_oci_auth(
                auth_type=kwargs.get("auth_type", "user_principal"),
                config_path=kwargs.get("config_path", oci.config.DEFAULT_LOCATION),
                profile=kwargs.get("profile", oci.config.DEFAULT_PROFILE),
                token=kwargs.get("token"),
                region=kwargs.get("region"),
            )
            return OCIModelAuthAdapter(oci_auth)

        elif provider == "aws-bedrock":
            return AWSBedrockAuth(
                access_key_id=kwargs.get("access_key_id"),
                secret_access_key=kwargs.get("secret_access_key"),
                session_token=kwargs.get("session_token"),
                region=kwargs.get("region"),
                profile=kwargs.get("profile"),
            )

        elif provider == "azure-openai":
            return AzureOpenAIAuth(
                api_key=kwargs.get("api_key"),
                api_version=kwargs.get("api_version", "2024-02-01"),  # Provide default
                azure_endpoint=kwargs.get("azure_endpoint"),
                azure_deployment=kwargs.get("azure_deployment"),
                use_azure_ad=kwargs.get("use_azure_ad", False),
                azure_ad_token=kwargs.get("azure_ad_token"),
            )

        elif provider == "gcp-vertex":
            return GCPVertexAuth(
                project_id=kwargs.get("project_id"),
                location=kwargs.get("location"),
                credentials_path=kwargs.get("credentials_path"),
                api_key=kwargs.get("api_key"),
            )

        elif provider == "together":
            api_key = kwargs.get("api_key")
            together_auth = TogetherAuth(api_key=api_key)
            return TogetherModelAuthAdapter(together_auth)

        else:
            raise ValueError(
                f"Unsupported model provider: {provider}. "
                f"Supported: openai, oci, aws-bedrock, azure-openai, gcp-vertex, "
                "together"
            )

    @staticmethod
    def create_storage_auth(provider: str, **kwargs) -> StorageAuthProvider:
        """Create a storage authentication provider.

        Args:
            provider: Provider type ('oci', 'aws', 'azure', 'gcp', 'github')
            **kwargs: Provider-specific arguments

        Returns:
            StorageAuthProvider instance

        Raises:
            ValueError: If provider is not supported
        """
        if provider == "oci":
            # Use existing OCI auth factory
            oci_auth = AuthFactory.create_oci_auth(
                auth_type=kwargs.get("auth_type", "user_principal"),
                config_path=kwargs.get("config_path", oci.config.DEFAULT_LOCATION),
                profile=kwargs.get("profile", oci.config.DEFAULT_PROFILE),
                token=kwargs.get("token"),
                region=kwargs.get("region"),
            )
            return OCIStorageAuthAdapter(oci_auth)

        elif provider == "aws":
            return AWSS3Auth(
                access_key_id=kwargs.get("access_key_id"),
                secret_access_key=kwargs.get("secret_access_key"),
                session_token=kwargs.get("session_token"),
                region=kwargs.get("region"),
                profile=kwargs.get("profile"),
            )

        elif provider == "azure":
            return AzureBlobAuth(
                account_name=kwargs.get("account_name"),
                account_key=kwargs.get("account_key"),
                connection_string=kwargs.get("connection_string"),
                sas_token=kwargs.get("sas_token"),
                use_azure_ad=kwargs.get("use_azure_ad", False),
                tenant_id=kwargs.get("tenant_id"),
                client_id=kwargs.get("client_id"),
                client_secret=kwargs.get("client_secret"),
            )

        elif provider == "gcp":
            return GCPStorageAuth(
                project_id=kwargs.get("project_id"),
                credentials_path=kwargs.get("credentials_path"),
                access_token=kwargs.get("access_token"),
            )

        elif provider == "github":
            return GitHubAuth(
                token=kwargs.get("token"),
                owner=kwargs.get("owner"),
                repo=kwargs.get("repo"),
            )

        else:
            raise ValueError(
                f"Unsupported storage provider: {provider}. "
                f"Supported: oci, aws, azure, gcp, github"
            )



from typing import Any, Dict

import oci

from genai_bench.auth.auth_provider import AuthProvider


class OCIOBOTokenAuth(AuthProvider):
    """OCI Authentication Provider using OBO (On-Behalf-Of) token."""

    def __init__(self, token: str, region: str):
        """Initialize OCI OBO Token Auth Provider.

        Args:
            token (str): The security token for OBO authentication
            region (str): The OCI region (e.g., 'us-ashburn-1')
        """
        self.token = token
        self.region = region
        self._signer = None

    def get_config(self) -> Dict[str, Any]:
        """Get OCI configuration.

        Returns:
            Dict[str, Any]: OCI configuration dictionary with region
        """
        return {
            "region": self.region,
        }

    def get_credentials(self) -> oci.signer.AbstractBaseSigner:
        """Get OCI security token signer.

        Returns:
            oci.signer.AbstractBaseSigner: OCI security token signer

        Raises:
            Exception: If token is invalid
        """
        if self._signer is None:
            self._signer = oci.auth.signers.SecurityTokenSigner(
                self.token, region=self.region
            )
        return self._signer



"""Factory for creating appropriate data loaders and loading data."""

from typing import Any, List, Tuple, Union, cast

from genai_bench.data.config import DatasetConfig
from genai_bench.data.loaders.image import ImageDatasetLoader
from genai_bench.data.loaders.text import TextDatasetLoader
from genai_bench.logging import init_logger

logger = init_logger(__name__)


class DataLoaderFactory:
    """Factory for creating data loaders and loading data."""

    @staticmethod
    def load_data_for_task(
        task: str, dataset_config: DatasetConfig
    ) -> Union[List[str], List[Tuple[str, Any]]]:
        """Load data for a specific task.

        Args:
            task: Task name in format "input-to-output"
            dataset_config: Dataset configuration

        Returns:
            Loaded data for the task
        """
        input_modality, output_modality = task.split("-to-")

        if input_modality == "text":
            return DataLoaderFactory._load_text_data(dataset_config, output_modality)
        elif "image" in input_modality:
            return DataLoaderFactory._load_image_data(dataset_config)
        else:
            raise ValueError(f"Unsupported input modality: {input_modality}")

    @staticmethod
    def _load_text_data(
        dataset_config: DatasetConfig, output_modality: str
    ) -> List[str]:
        """Load text data."""

        loader = TextDatasetLoader(dataset_config)
        data = loader.load_request()

        # TextDatasetLoader always returns List[str]
        text_data = cast(List[str], data)
        return text_data

    @staticmethod
    def _load_image_data(
        dataset_config: DatasetConfig,
    ) -> List[Tuple[str, Any]]:
        """Load image data."""
        loader = ImageDatasetLoader(dataset_config)
        data = loader.load_request()
        return data  # type: ignore



import json
import os
from pathlib import Path

import click
from huggingface_hub.utils import HfHubHTTPError
from transformers import AutoTokenizer

from genai_bench.data.config import DatasetConfig
from genai_bench.logging import init_logger
from genai_bench.scenarios.base import Scenario
from genai_bench.user.aws_bedrock_user import AWSBedrockUser
from genai_bench.user.azure_openai_user import AzureOpenAIUser
from genai_bench.user.cohere_user import CohereUser
from genai_bench.user.gcp_vertex_user import GCPVertexUser
from genai_bench.user.oci_cohere_user import OCICohereUser
from genai_bench.user.oci_genai_user import OCIGenAIUser
from genai_bench.user.openai_user import OpenAIUser
from genai_bench.user.together_user import TogetherUser

logger = init_logger(__name__)

API_BACKEND_USER_MAP = {
    OpenAIUser.BACKEND_NAME: OpenAIUser,
    OCICohereUser.BACKEND_NAME: OCICohereUser,
    OCIGenAIUser.BACKEND_NAME: OCIGenAIUser,
    CohereUser.BACKEND_NAME: CohereUser,
    AWSBedrockUser.BACKEND_NAME: AWSBedrockUser,
    AzureOpenAIUser.BACKEND_NAME: AzureOpenAIUser,
    GCPVertexUser.BACKEND_NAME: GCPVertexUser,
    TogetherUser.BACKEND_NAME: TogetherUser,
    "vllm": OpenAIUser,  # vLLM uses OpenAI-compatible API
    "sglang": OpenAIUser,  # SGLang uses OpenAI-compatible API
}

DEFAULT_NUM_CONCURRENCIES = [1, 2, 4, 8, 16, 32, 64, 128, 256]
DEFAULT_BATCH_SIZES = [1, 2, 4, 8, 24, 48, 96]

DEFAULT_SCENARIOS_FOR_CHAT = [
    "N(480,240)/(300,150)",
    "D(100,100)",
    "D(100,1000)",
    "D(2000,200)",
    "D(7800,200)",
]


DEFAULT_SCENARIOS_FOR_VISION = [
    "I(512,512)",
    "I(1024,512)",
    "I(2048,2048)",
]

DEFAULT_SCENARIOS_FOR_EMBEDDING = [
    "E(64)",
    "E(128)",
    "E(256)",
    "E(512)",
    "E(1024)",
]

DEFAULT_SCENARIOS_FOR_RERANK = [
    "R(64,100)",
    "R(128,100)",
    "R(256,100)",
    "R(512,100)",
    "R(1024,100)",
    "R(2096,100)",
    "R(4096,100)",
]

DEFAULT_SCENARIOS_BY_TASK = {
    "text-to-text": DEFAULT_SCENARIOS_FOR_CHAT,
    "text-to-rerank": DEFAULT_SCENARIOS_FOR_RERANK,
    "image-text-to-text": DEFAULT_SCENARIOS_FOR_VISION,
    "text-to-embeddings": DEFAULT_SCENARIOS_FOR_EMBEDDING,
    "image-to-embeddings": DEFAULT_SCENARIOS_FOR_VISION,
    # add other tasks and default scenarios as needed
}


# -------------------------
# General Validation Functions
# -------------------------
def validate_dataset_path_callback(ctx, param, value):
    """Validate the dataset path."""
    task = ctx.params.get("task")
    if not task:
        raise click.BadParameter(
            "The '--task' option is required but was not provided."
        )

    input_modality, output_modality = task.split("-to-")
    if "image" in input_modality and value is None:
        # Check if dataset_config is provided as alternative
        dataset_config = ctx.params.get("dataset_config")
        if dataset_config is None:
            raise click.BadParameter(
                '--dataset-path is required when --task includes "image" input '
                "modality and --dataset-config is not provided."
            )
        else:
            logger.warning(
                "Using dataset configuration file for image task. "
                "Ensure your config file specifies the correct image dataset source."
            )
    return value


def validate_dataset_config(ctx, param, value):
    """Validate dataset configuration file."""
    if value is None:
        return None

    try:
        _ = DatasetConfig.from_file(value)
        return value
    except Exception as e:
        raise click.BadParameter(f"Invalid dataset configuration file: {str(e)}") from e


def validate_scenario_callback(scenario):
    """Validate an individual traffic scenario."""
    try:
        Scenario.validate(scenario)
        return scenario
    except ValueError as e:
        raise click.BadParameter(str(e)) from e


def validate_traffic_scenario_callback(ctx, param, value):
    """Validate and assign traffic scenarios based on the task."""
    task = ctx.params.get("task")
    if not task:
        raise click.BadParameter(
            "The '--task' option is required but was not provided."
        )
    if value:
        return [validate_scenario_callback(v) for v in value]
    # If user provided a dataset and no scenario, prefer dataset mode (no data shaping)
    if ctx.params.get("dataset_path") or ctx.params.get("dataset_config"):
        return ["dataset"]
    if task not in DEFAULT_SCENARIOS_BY_TASK:
        raise click.BadParameter(
            f"No default traffic scenarios defined for task '{task}'"
        )

    logger.info(f"Using default traffic scenarios for task {task}")
    return DEFAULT_SCENARIOS_BY_TASK[task]


def validate_tokenizer(model_tokenizer):
    """Validate and load the tokenizer, either locally or from HuggingFace.

    The function now tries an **anonymous** download first to support public
    repositories without requiring an ``HF_TOKEN``. Authentication is only
    enforced when the Hugging Face Hub returns *401* or *403* errors.
    """
    if isinstance(model_tokenizer, str) and Path(model_tokenizer).exists():
        return AutoTokenizer.from_pretrained(model_tokenizer)
    hf_token = os.environ.get("HF_TOKEN")

    try:
        return AutoTokenizer.from_pretrained(model_tokenizer, token=hf_token)
    except (HfHubHTTPError, OSError) as e:
        is_auth_error = False

        if isinstance(e, OSError):
            if (
                "gated repo" in str(e)
                or "private repository" in str(e)
                and "make sure to pass a token" in str(e)
            ):
                is_auth_error = True
        elif isinstance(e, HfHubHTTPError) and e.response is not None:
            is_auth_error = e.response.status_code in {401, 403, 404}

        if is_auth_error and hf_token is None:
            raise click.BadParameter(
                "Hugging Face requires authentication for this tokenizer. "
                "Please export HF_TOKEN with a valid access token and retry."
            ) from e

        # Propagate all other errors unchanged
        raise


def validate_iteration_params(ctx, param, value) -> str:
    """
    Validate and determine iteration parameters based on task type.

    Args:
        ctx: Click context
        param: Click parameter
        value: Current parameter value (iteration_type)

    Returns:
        str: The validated iteration_type
    """
    task = ctx.params.get("task")
    num_concurrency = ctx.params.get("num_concurrency", [])
    batch_size = ctx.params.get("batch_size", [])

    # For text-to-embeddings tasks, always use batch_size iteration
    if task == "text-to-embeddings" or task == "text-to-rerank":
        if value != "batch_size":
            click.echo("Note: Using batch_size iteration for text-to-embeddings task")
        batch_size = batch_size or DEFAULT_BATCH_SIZES
        value = "batch_size"
        num_concurrency = [1]

    # For all other tasks, use num_concurrency iteration
    else:
        if value != "num_concurrency":
            click.echo(f"Note: Using num_concurrency iteration for {task} task")
        num_concurrency = num_concurrency or DEFAULT_NUM_CONCURRENCIES
        value = "num_concurrency"
        batch_size = [1]

    # Update context with validated values
    ctx.params.update(
        {
            "iteration_type": value,
            "batch_size": batch_size,
            "num_concurrency": num_concurrency,
        }
    )

    return value


# -------------------------
# CLI Validation Functions
# -------------------------


def validate_api_backend(ctx, param, value):
    """
    Validate the selected API backend and set the corresponding user class.
    """
    api_backend = value.lower()
    user_class = API_BACKEND_USER_MAP.get(api_backend)
    if not user_class:
        raise click.BadParameter(f"{value} is not a supported API backend.")

    if ctx.obj is None:
        ctx.obj = {}
    ctx.obj["user_class"] = user_class

    return api_backend


def validate_api_key(ctx, param, value):
    """Validate API key based on backend."""
    api_backend = ctx.params.get("api_backend")

    if not api_backend:
        raise click.BadParameter("api_backend must be specified before api_key")

    # Backends that require API key
    api_key_required = [OpenAIUser.BACKEND_NAME, "vllm", "sglang"]

    # Backends that don't use traditional API key
    no_api_key = [
        OCICohereUser.BACKEND_NAME,
        OCIGenAIUser.BACKEND_NAME,
        CohereUser.BACKEND_NAME,
        AWSBedrockUser.BACKEND_NAME,
        GCPVertexUser.BACKEND_NAME,
    ]

    # Azure OpenAI can use API key or Azure AD
    if api_backend in api_key_required:
        if not value:
            raise click.BadParameter(f"API key is required for {api_backend} backend")
    elif api_backend == AzureOpenAIUser.BACKEND_NAME:
        # Azure can use API key or Azure AD - validated in model auth options
        pass
    elif api_backend in no_api_key:
        if value:
            click.echo(
                f"Warning: API key is not used for {api_backend} backend "
                f"as it uses cloud-specific authentication",
                err=True,
            )
        return None

    return value


def validate_task(ctx, param, value):
    """
    Validate the selected task and ensure compatibility with the API backend.
    """
    task = value.lower()

    # Ensure API backend has been validated
    user_class = ctx.obj.get("user_class") if ctx.obj else None
    if not user_class:
        raise click.BadParameter(
            "API backend is not set. Please provide a valid --api-backend "
            "before --task."
        )

    # Check task compatibility
    if not user_class.is_task_supported(task):
        supported_tasks = ", ".join(user_class.supported_tasks.keys())
        raise click.BadParameter(
            f"Task '{task}' is not supported by the selected API backend. "
            f"Supported tasks are: {supported_tasks}."
        )

    # Store the user task function in the context
    ctx.obj["user_task"] = getattr(user_class, user_class.supported_tasks[task])

    return task


def set_model_from_tokenizer(ctx, param, value):
    """
    Infer the model name from the tokenizer if not explicitly provided.
    """
    model_tokenizer = ctx.params.get("model_tokenizer")
    return value or model_tokenizer.split("/")[-1]


def validate_additional_request_params(ctx, param, value):
    """Validate and parse additional request parameters in JSON format."""
    if value is None:
        return {}  # Return an empty dictionary to be parsed in Sampler
    try:
        value = json.loads(value)
        if "temperature" in value and value["temperature"] > 1.5:
            warning_msg = (
                f"You have set temperature {value['temperature']} too high. "
                f"This may cause higher chars_to_token ratio in the metrics "
                f"and result in higher total_chars_per_hour."
            )
            logger.warning(warning_msg)
            click.secho(warning_msg, fg="yellow", bold=True)
            if click.confirm("Do you want to re-enter a temperature?", default=False):
                value["temperature"] = click.prompt(
                    "Please enter a new temperature,",
                    default=0,
                )
        if "ignore_eos" in value and not value["ignore_eos"]:
            warning_msg = (
                "You have set ignore_eos to False. This will cause inaccurate "
                "num_output_tokens because the generation can stop before "
                "reaching the num_output_tokens you passed. Unless you are "
                "doing correctness test, it is recommended to set ignore_eos "
                "to True."
            )
            logger.warning(warning_msg)
            click.secho(warning_msg, fg="yellow", bold=True)
            if click.confirm(
                "Do you want to reset the ignore_eos to True?", default=False
            ):
                value["ignore_eos"] = True

        return value
    except json.JSONDecodeError as e:
        raise click.BadParameter(
            f"Invalid JSON string in `--additional-request-params`: {str(e)}"
        ) from e


def validate_filter_criteria(ctx, param, value):
    """Parse and validate the filter criteria provided by the user."""
    if value is None:
        return {}
    try:
        return json.loads(value)
    except json.JSONDecodeError as e:
        raise click.BadParameter(
            f"Invalid JSON string in `--filter-criteria`: {str(e)}"
        ) from e


def validate_object_storage_options(ctx, param, value):
    """Validate object storage options."""
    if (
        param.name == "upload_results"
        and value
        and not ctx.params.get("storage_bucket")
    ):
        raise click.UsageError(
            "You must provide a storage bucket name (--storage-bucket) "
            "when uploading results."
        )
    return value


def validate_warmup_cooldown_ratio_options(ctx, param, value):
    """Validate warmup and cooldown options."""
    warmup_ratio = ctx.params.get("warmup_ratio") or 0.0
    cooldown_ratio = value or 0.0
    if warmup_ratio + cooldown_ratio >= 1.0:
        raise click.BadParameter(
            f"warmup_ratio({warmup_ratio}) + cooldown_ratio({cooldown_ratio}) "
            f"must be < 1.0.",
            param_hint=["--warmup-ratio", "--cooldown-ratio"],
        )
    return value
